from typing import TYPE_CHECKING, Annotated, Iterable

import strawberry
from strawberry import UNSET
from strawberry.federation.schema_directives import Key

from oauth2_lib.strawberry import authenticated_field
from orchestrator.db import ProductBlockTable, ProductTable
from orchestrator.domain.base import ProductModel
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.schemas.fixed_input import FixedInput
from orchestrator.graphql.schemas.helpers import get_original_model
from orchestrator.graphql.schemas.product_block import ProductBlock
from orchestrator.graphql.schemas.workflow import Workflow
from orchestrator.graphql.types import GraphqlFilter, GraphqlSort, OrchestratorInfo
from orchestrator.schemas.product import ProductSchema

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.subscription import SubscriptionInterface


federation_key_directives = [Key(fields="productId", resolvable=UNSET)]


@strawberry.experimental.pydantic.type(model=ProductSchema, directives=federation_key_directives)
class ProductType:
    product_id: strawberry.auto
    name: strawberry.auto
    description: strawberry.auto
    product_type: strawberry.auto
    status: strawberry.auto
    tag: strawberry.auto
    created_at: strawberry.auto
    end_date: strawberry.auto

    @strawberry.field(description="Returns the product type")  # type: ignore
    async def type(self) -> str:
        return self.product_type

    @authenticated_field(description="Returns list of subscriptions of the product type")  # type: ignore
    async def subscriptions(
        self,
        info: OrchestratorInfo,
        filter_by: list[GraphqlFilter] | None = None,
        sort_by: list[GraphqlSort] | None = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[Annotated["SubscriptionInterface", strawberry.lazy(".subscription")]]:
        from orchestrator.graphql.resolvers.subscription import resolve_subscriptions

        filter_by_with_related_subscriptions = (filter_by or []) + [GraphqlFilter(field="product", value=self.name)]
        return await resolve_subscriptions(info, filter_by_with_related_subscriptions, sort_by, first, after)

    @strawberry.field(description="Returns list of all nested productblock names")  # type: ignore
    async def all_pb_names(self) -> list[str]:

        model = get_original_model(self, ProductTable)

        def get_all_pb_names(product_blocks: list[ProductBlockTable]) -> Iterable[str]:
            for product_block in product_blocks:
                yield product_block.name

                if product_block.depends_on:
                    yield from get_all_pb_names(product_block.depends_on)

        names: list[str] = list(get_all_pb_names(model.product_blocks))
        names.sort()

        return names

    @strawberry.field(description="Return product blocks")  # type: ignore
    async def product_blocks(self) -> list[Annotated["ProductBlock", strawberry.lazy(".product_block")]]:
        from orchestrator.graphql.schemas.product_block import ProductBlock

        model = get_original_model(self, ProductTable)

        return [ProductBlock.from_pydantic(i) for i in model.product_blocks]

    @strawberry.field(description="Return fixed inputs")  # type: ignore
    async def fixed_inputs(self) -> list[Annotated["FixedInput", strawberry.lazy(".fixed_input")]]:
        from orchestrator.graphql.schemas.fixed_input import FixedInput

        model = get_original_model(self, ProductTable)

        return [FixedInput.from_pydantic(i) for i in model.fixed_inputs]

    @strawberry.field(description="Return workflows")  # type: ignore
    async def workflows(self) -> list[Annotated["Workflow", strawberry.lazy(".workflow")]]:
        from orchestrator.graphql.schemas.workflow import Workflow

        model = get_original_model(self, ProductTable)

        return [Workflow.from_pydantic(i) for i in model.workflows]


@strawberry.experimental.pydantic.type(model=ProductModel, all_fields=True)
class ProductModelGraphql:
    @strawberry.field(description="Returns the product type")  # type: ignore
    async def type(self) -> str:
        return self.product_type  # type: ignore



from enum import IntEnum

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from pydantic import computed_field


class ExampleIntEnum2(IntEnum):
    _1 = 1
    _2 = 2
    _3 = 3
    _4 = 4


class Example2BlockInactive(ProductBlockModel, product_block_name="Example2"):
    example_int_enum_2: ExampleIntEnum2 | None = None


class Example2BlockProvisioning(Example2BlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    example_int_enum_2: ExampleIntEnum2 | None = None

    @computed_field
    @property
    def title(self) -> str:
        # TODO: format correct title string
        return f"{self.name}"


class Example2Block(Example2BlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    example_int_enum_2: ExampleIntEnum2



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ipaddress
import re
from dataclasses import dataclass
from datetime import datetime, timezone
from uuid import UUID

import pytest
import pytz
from pydantic import BaseModel

from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.utils.datetime import nowtz
from orchestrator.utils.json import from_serializable, json_dumps, json_loads, non_none_dict, to_serializable


def test_serialization_datetime():
    json_str = json_dumps({"end_date": nowtz()})
    assert re.search(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+00:00", json_str)


def test_deserialization_datetime():
    json_str = '{"end_date": "2019-12-06T19:25:22+00:00"}'
    dct = json_loads(json_str)
    assert "end_date" in dct
    assert dct["end_date"] == datetime(2019, 12, 6, 19, 25, 22, 0, timezone.utc)

    dct = {"end_date": datetime(2019, 12, 6, 19, 25, 22, 0, timezone.utc)}
    assert json_loads(json_dumps(dct)) == dct


def test_no_none_dict():
    assert non_none_dict({}.items()) == {}
    assert non_none_dict({"a": 0, "b": None, "c": ""}.items()) == {"a": 0, "c": ""}


def test_to_serializable():
    assert to_serializable(UUID("3ec5fad4-e58c-4cb3-a860-6764cfc658f3")) == "3ec5fad4-e58c-4cb3-a860-6764cfc658f3"
    assert to_serializable(ipaddress.ip_address("10.0.0.1")) == "10.0.0.1"
    assert to_serializable(ipaddress.ip_address("fd00::1")) == "fd00::1"
    assert to_serializable(ipaddress.ip_network("10.0.0.0/24")) == "10.0.0.0/24"
    assert to_serializable(ipaddress.ip_network("fd00::0/64")) == "fd00::/64"
    assert to_serializable(datetime(2021, 7, 28, 1, 1, 1)) == "2021-07-28T01:01:01"
    assert to_serializable({3, 4, 5, 5, 6}) == [3, 4, 5, 6]

    @dataclass
    class Foo:
        bar: int

    assert to_serializable(Foo(1)) == {"bar": 1}

    class Foo:
        def __json__(self):
            return {"bar": 1}

    assert to_serializable(Foo()) == {"bar": 1}

    class Foo:
        def to_dict(self):
            return {"bar": 1}

    assert to_serializable(Foo()) == {"bar": 1}

    class Foo(BaseModel):
        bar: int

    assert to_serializable(Foo(bar=1)) == {"bar": 1}

    with pytest.raises(TypeError, match=r"Could not serialize object of type int to JSON"):
        to_serializable(1)


def test_from_serializable():
    assert from_serializable(
        {"date": "2021-07-28T01:01:01+00:00", "date-to-short": "2021-07-28T01:01:01", "int": 1}
    ) == {
        "date": datetime(2021, 7, 28, 1, 1, 1, tzinfo=pytz.UTC),
        "date-to-short": "2021-07-28T01:01:01",
        "int": 1,
    }


def test_orchestrator_base_serializer():

    class Foo(OrchestratorBaseModel):
        baz: datetime
        bar: int

    foo = Foo(baz=datetime(2024, 12, 1, 1, 1), bar=1)
    assert foo.model_dump(mode="json") == {"baz": datetime(2024, 12, 1, 1, 1).timestamp(), "bar": 1}


def test_orchestrator_base_serializer_recursive():
    class Bar(OrchestratorBaseModel):
        baz: datetime

    class Foo(OrchestratorBaseModel):
        baz: datetime
        nested_bar: Bar
        bing: int

    foo = Foo(baz=datetime(2024, 12, 1, 1, 1), nested_bar=Bar(baz=datetime(2024, 11, 1, 1, 1)), bing=1)
    assert foo.model_dump(mode="json") == {
        "baz": datetime(2024, 12, 1, 1, 1).timestamp(),
        "nested_bar": {"baz": datetime(2024, 11, 1, 1, 1).timestamp()},
        "bing": 1,
    }



from http import HTTPStatus
from unittest import mock
from unittest.mock import Mock

from sqlalchemy.exc import SQLAlchemyError

from orchestrator.db import db
from orchestrator.services.settings import get_engine_settings


def test_get_engine_status(test_client):
    engine_settings = get_engine_settings()
    response = test_client.get("/api/settings/status")
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 0
    assert response.json()["global_lock"] is False
    assert response.json()["global_status"] == "RUNNING"

    engine_settings.global_lock = True
    engine_settings.running_processes = 1
    db.session.flush()

    response = test_client.get("/api/settings/status")
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 1
    assert response.json()["global_lock"] is True
    assert response.json()["global_status"] == "PAUSING"

    engine_settings.running_processes = 0
    db.session.flush()

    response = test_client.get("/api/settings/status")
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 0
    assert response.json()["global_lock"] is True
    assert response.json()["global_status"] == "PAUSED"


def test_set_engine_status(test_client):
    response = test_client.put("/api/settings/status", json={"global_lock": True})
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 0
    assert response.json()["global_lock"] is True
    assert response.json()["global_status"] == "PAUSED"

    response = test_client.put("/api/settings/status", json={"global_lock": False})
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 0
    assert response.json()["global_lock"] is False
    assert response.json()["global_status"] == "RUNNING"


def test_reset_search_index(test_client, generic_subscription_1, generic_subscription_2):
    # Initially no subscriptions found in index
    response = test_client.get("/api/subscriptions/search?query=status:active")
    assert response.status_code == HTTPStatus.OK
    subscriptions = response.json()
    assert len(subscriptions) == 0

    # Refresh index
    response = test_client.post("/api/settings/search-index/reset")
    assert response.status_code == HTTPStatus.OK

    # Subscriptions are now found
    response = test_client.get("/api/subscriptions/search?query=status:active")
    subscriptions = response.json()

    assert len(subscriptions) == 2


def test_reset_search_index_error(test_client, generic_subscription_1, generic_subscription_2):
    with mock.patch.object(db, "session") as ex:
        session_execute_mock = Mock(side_effect=SQLAlchemyError("Database error"))
        ex.attach_mock(session_execute_mock, "execute")
        response = test_client.post("/api/settings/search-index/reset")
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR



from uuid import uuid4

from orchestrator.forms.validators import DisplaySubscription, Label, migration_summary
from pydantic_forms.core import FormPage


def test_display_subscription():
    some_sub_id = uuid4()

    class Form(FormPage):
        display_sub: DisplaySubscription = some_sub_id

    expected = {"display_sub": some_sub_id}

    assert Form().model_dump() == expected


def test_display_subscription_update_not_allowed():
    some_sub_id = uuid4()

    class Form(FormPage):
        display_sub: DisplaySubscription = some_sub_id

    expected = {"display_sub": some_sub_id}

    assert Form(display_sub=uuid4()).model_dump() == expected


def test_display_only_schema():
    some_sub_id = uuid4()
    Summary = migration_summary({"headers": ["one"]})

    class Form(FormPage):
        display_sub: DisplaySubscription = some_sub_id
        label: Label
        summary: Summary

    expected = {
        "$defs": {
            "MigrationSummaryValue": {
                "properties": {},
                "title": "MigrationSummaryValue",
                "type": "object",
            },
        },
        "additionalProperties": False,
        "properties": {
            "display_sub": {
                "default": str(some_sub_id),
                "format": "subscription",
                "title": "Display Sub",
                "type": "string",
            },
            "label": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "format": "label",
                "default": None,
                "title": "Label",
                "type": "string",
            },
            "summary": {
                "allOf": [
                    {
                        "$ref": "#/$defs/MigrationSummaryValue",
                    },
                ],
                "format": "summary",
                "default": None,
                "type": "string",
                "uniforms": {"data": {"headers": ["one"]}},
            },
        },
        "title": "unknown",
        "type": "object",
    }

    assert Form.model_json_schema() == expected



from uuid import UUID, uuid4

import pytest
from sqlalchemy import func, select

from orchestrator.db import SubscriptionTable, db
from orchestrator.domain import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum


@pytest.fixture
def create_vertical_subscription(
    test_product_model_nested, test_product_type_one_nested, test_product_block_one_nested
):
    ProductTypeOneNestedForTestInactive, _, _ = test_product_type_one_nested
    ProductBlockOneNestedForTestInactive, _, _ = test_product_block_one_nested

    customer_id = str(uuid4())

    def create(*, size: int, depth: int = 1) -> UUID:
        if depth == size:
            sub_block = None
        else:
            nested_subscription_id = create(size=size, depth=depth + 1)
            nested_subscription = ProductTypeOneNestedForTestInactive.from_subscription(nested_subscription_id)
            sub_block = nested_subscription.block

        subscription = ProductTypeOneNestedForTestInactive.from_product_id(
            product_id=test_product_model_nested.product_id, customer_id=customer_id, insync=True
        )
        subscription.block = ProductBlockOneNestedForTestInactive.new(
            subscription_id=subscription.subscription_id, int_field=depth, sub_block=sub_block
        )
        subscription.save()

        return subscription.subscription_id

    yield create


@pytest.fixture
def create_horizontal_subscription(
    test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one

    customer_id = str(uuid4())

    def create(*, size: int) -> UUID:
        subscription = ProductTypeOneForTestInactive.from_product_id(
            product_id=test_product_model.product_id, customer_id=customer_id, insync=True
        )

        def create_subblock(value):
            return SubBlockOneForTestInactive.new(subscription.subscription_id, int_field=value, str_field=str(value))

        subscription.block = ProductBlockOneForTestInactive.new(
            subscription_id=subscription.subscription_id,
            int_field=123,
            str_field="abc",
            sub_block=create_subblock(1),
            sub_block_2=create_subblock(2),
            sub_block_list=[create_subblock(n) for n in range(10, 10 + size)],
            enum_field=DummyEnum.FOO,
        )
        subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.ACTIVE)
        subscription.save()

        return subscription.subscription_id

    yield create


@pytest.fixture
def subscription_with_100_horizontal_blocks(create_horizontal_subscription):
    return create_horizontal_subscription(size=100)


@pytest.mark.benchmark
def test_subscription_model_horizontal_references(
    subscription_with_100_horizontal_blocks, test_product_type_one, monitor_sqlalchemy
):
    # Note: fixtures only execute once per benchmark and are excluded from the measurement

    # given
    _, _, ProductTypeOneForTest = test_product_type_one

    subscription_id = subscription_with_100_horizontal_blocks
    db.session.expunge_all()  # otherwise sqlalchemy will just serve everything from cache

    # when

    with monitor_sqlalchemy():  # Context does nothing unless you set CLI_OPT_MONITOR_SQLALCHEMY
        subscription = ProductTypeOneForTest.from_subscription(subscription_id)

    # then
    assert len(subscription.block.sub_block_list) == 100


@pytest.fixture
def subscription_with_10_vertical_blocks(create_vertical_subscription):
    return create_vertical_subscription(size=10)


@pytest.mark.benchmark
def test_subscription_model_vertical_references(
    subscription_with_10_vertical_blocks, test_product_type_one_nested, monitor_sqlalchemy
):
    # Note: fixtures only execute once per benchmark and are excluded from the measurement

    # given
    _, _, ProductTypeOneNestedForTest = test_product_type_one_nested

    subscription_id = subscription_with_10_vertical_blocks
    db.session.expunge_all()  # otherwise sqlalchemy will just serve everything from cache

    # when

    with monitor_sqlalchemy():  # Context does nothing unless you set CLI_OPT_MONITOR_SQLALCHEMY
        subscription = ProductTypeOneNestedForTest.from_subscription(subscription_id)

    # then
    assert subscription.block is not None
    assert subscription.block.sub_block is not None
    assert subscription.block.sub_block.sub_block is not None
    assert subscription.block.sub_block.sub_block.sub_block is not None
    # no need to check all x levels


@pytest.mark.benchmark
def test_subscription_model_vertical_references_save(create_vertical_subscription, monitor_sqlalchemy):
    # when
    with monitor_sqlalchemy():
        subscription_id = create_vertical_subscription(size=5)

    # then

    # Checks that the subscription was created, without too much overhead
    query_check_created = (
        select(func.count()).select_from(SubscriptionTable).where(SubscriptionTable.subscription_id == subscription_id)
    )
    assert db.session.scalar(query_check_created) == 1


@pytest.mark.benchmark
def test_subscription_model_horizontal_references_save(create_horizontal_subscription, monitor_sqlalchemy):
    # when
    with monitor_sqlalchemy():
        subscription_id = create_horizontal_subscription(size=10)

    # then

    # Checks that the subscription was created, without too much overhead
    query_check_created = (
        select(func.count()).select_from(SubscriptionTable).where(SubscriptionTable.subscription_id == subscription_id)
    )
    assert db.session.scalar(query_check_created) == 1



import structlog
from orchestrator.domain import SubscriptionModel
from orchestrator.forms import FormPage
from orchestrator.forms.validators import CustomerId, Divider
from orchestrator.types import SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.steps import set_status
from orchestrator.workflows.utils import modify_workflow
from pydantic_forms.types import FormGenerator, State, UUIDstr

from products.product_blocks.example2 import ExampleIntEnum2
from products.product_types.example2 import Example2, Example2Provisioning


def subscription_description(subscription: SubscriptionModel) -> str:
    """The suggested pattern is to implement a subscription service that generates a subscription specific
    description, in case that is not present the description will just be set to the product name.
    """
    return f"{subscription.product.name} subscription"


logger = structlog.get_logger(__name__)


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    subscription = Example2.from_subscription(subscription_id)
    example2 = subscription.example2

    # TODO fill in additional fields if needed

    class ModifyExample2Form(FormPage):
        customer_id: CustomerId = subscription.customer_id  # type: ignore

        divider_1: Divider

        example_int_enum_2: ExampleIntEnum2 | None = example2.example_int_enum_2

    user_input = yield ModifyExample2Form
    user_input_dict = user_input.dict()

    return user_input_dict | {"subscription": subscription}


@step("Update subscription")
def update_subscription(
    subscription: Example2Provisioning,
    example_int_enum_2: ExampleIntEnum2 | None,
) -> State:
    # TODO: get all modified fields
    subscription.example2.example_int_enum_2 = example_int_enum_2

    return {"subscription": subscription}


@step("Update subscription description")
def update_subscription_description(subscription: Example2) -> State:
    subscription.description = subscription_description(subscription)
    return {"subscription": subscription}


additional_steps = begin


@modify_workflow("Modify example2", initial_input_form=initial_input_form_generator, additional_steps=additional_steps)
def modify_example2() -> StepList:
    return (
        begin
        >> set_status(SubscriptionLifecycle.PROVISIONING)
        >> update_subscription
        >> update_subscription_description
        # TODO add additional steps if needed
        >> set_status(SubscriptionLifecycle.ACTIVE)
    )



import re
from difflib import context_diff
from filecmp import dircmp
from pathlib import Path

import pytest
import structlog
from more_itertools import one

from orchestrator.cli.generate import app as generate_app
from test.unit_tests.cli.helpers import absolute_path

logger = structlog.get_logger()


@pytest.fixture(scope="module")
def actual_folder(cli_invoke, tmp_generate_path) -> Path:
    for config_file in (
        absolute_path("product_config2.yaml"),
        absolute_path("product_config1.yaml"),
        absolute_path("product_config4.yaml"),
    ):
        for cmd in ("product-blocks", "product", "workflows", "unit-tests"):
            cli_invoke(generate_app, [cmd, "--config-file", config_file, "--no-dryrun", "--force"])

        cli_invoke(generate_app, ["migration", "--config-file", config_file])

    return tmp_generate_path


@pytest.fixture(scope="module")
def expected_folder() -> Path:
    return Path(__file__).resolve().parent / "data" / "generate"


def test_missing_or_extra_files(expected_folder, actual_folder):
    def assert_equal_dirs(dcmp):
        left_only = ", ".join(dcmp.left_only)
        right_only = ", ".join(dcmp.right_only)
        relative_folder = Path(dcmp.right).relative_to(actual_folder)

        assert left_only == "", f"missing file(s) {left_only} from directory {relative_folder}"
        assert right_only == "", f"extra file(s) {right_only} in directory {relative_folder}"

        for sub_dcmp in dcmp.subdirs.values():
            assert_equal_dirs(sub_dcmp)

    assert_equal_dirs(dircmp(expected_folder, actual_folder, ignore=["versions", "__pycache__"]))  # skip migrations


def get_dynamic_migration_values(file):
    """Retrieve dynamic values from the given migration file."""
    line_prefixes = ["Revision ID:", "Revises:", "Create Date:", "revision =", "down_revision =", "depends_on ="]
    text = file.open().read()
    regex = r"^(?:(%s)(.*))$" % ("|".join(re.escape(v) for v in line_prefixes),)
    matches = re.findall(regex, text, flags=re.MULTILINE)
    return dict(matches)  # {"Revision ID:": "Revision ID: 59e1199aff7f", ...}


def update_dynamic_migration_values(file, replacements):
    """Update dynamic values in the migration file and return the content."""
    text = file.open().read()
    for prefix, replacement_line in replacements.items():
        regex_pattern = r"^(?:(%s)(.*))$" % (re.escape(prefix),)
        regex_replace = r"\1%s" % (replacement_line,)
        text = re.sub(regex_pattern, regex_replace, text, count=1, flags=re.MULTILINE)
    return text


def get_revision_ids(folder):
    """Find all alembic migrations and return a list of revision ids starting from the base.

    Note that this will only return the chain of revisions starting from the base revision.
    A warning will be logged in case of unconnected revisions because this is probably an error.
    """

    def showfile(file):
        return str(file.relative_to(folder))

    revision_chain = {}
    migration_files = list((folder / "migrations/versions/schema").glob("*.py"))
    rev_id_to_file = {}
    for migration_file in migration_files:
        _date, filename_rev_id, *_rest = migration_file.name.split("_")
        rev_ids = dict(re.findall(r'(.+) = "([0-9a-z]+)"', migration_file.open().read()))
        migration_rev_id = rev_ids["revision"]
        assert (
            filename_rev_id == migration_rev_id
        ), f"Migration file {showfile(migration_file)} has a different revision id in name and body"
        rev_id_to_file[migration_rev_id] = migration_file
        revision_chain[rev_ids.get("down_revision", "base")] = migration_rev_id

    def follow_rev(rev):
        if rev not in revision_chain:
            return
        yield revision_chain[rev]
        yield from follow_rev(revision_chain[rev])

    revision_history = list(follow_rev("base"))
    if len(migration_files) > len(revision_history):
        unconnected = [showfile(file) for rev_id, file in rev_id_to_file.items() if rev_id not in revision_history]
        logger.warning(
            "Not all migration files in this directory are connected to the chain of revisions",
            directory=folder,
            revision_chain=revision_history,
            unconnected=unconnected,
        )
    return revision_history


def get_expected_to_actual_migration_rev_ids(actual_folder, expected_folder):
    expected_revision_ids = get_revision_ids(expected_folder)
    actual_revision_ids = get_revision_ids(actual_folder)
    # If you're missing an expected migration, make sure it's down_revision correctly points to the previous one
    assert len(actual_revision_ids) == len(expected_revision_ids), (
        "Number of expected migration files does not match the number of actual migration files, "
        "check the logs above for hints"
    )
    return dict(zip(expected_revision_ids, actual_revision_ids))


def test_differences_in_generated_code(expected_folder, actual_folder):
    _expected_to_actual_rev_id = get_expected_to_actual_migration_rev_ids(actual_folder, expected_folder)

    def derive_actual_migration_file(expected_file):
        _expected_date, expected_rev_id, *rest = expected_file.name.split("_")
        actual_rev_id = _expected_to_actual_rev_id[expected_rev_id]
        return one((actual_folder / "migrations/versions/schema").glob(f"*_{actual_rev_id}_*.py"))

    for expected in expected_folder.rglob("*.py"):
        relative = expected.relative_to(expected_folder)
        actual = actual_folder / relative

        if not actual.exists():
            # Must be a migration file which is dynamic name (contains date and revision ID)
            actual = derive_actual_migration_file(expected)
            replacements = get_dynamic_migration_values(expected)
            tofile = f"actual {actual.relative_to(actual_folder)} (derived)"
            actual_lines = update_dynamic_migration_values(actual, replacements).splitlines(keepends=True)
        else:
            tofile = f"actual {relative}"
            actual_lines = actual.open().readlines()

        diff = context_diff(
            open(expected).readlines(), actual_lines, fromfile=f"expected {relative}", tofile=tofile, n=1, lineterm=""
        )

        formatted_diff = "\n".join(diff)
        if formatted_diff:
            print(f"\n{formatted_diff}")

        assert not formatted_diff, f"generated {relative} differs, see expected vs actual in the stdout output"



from test.unit_tests.conftest import CUSTOMER_ID


def test_process_metrics_success(test_client, mocked_processes) -> None:
    response = test_client.get("api/metrics")
    expected_metric_lines = [
        "# HELP wfo_process_count Number of processes per status, creator, task, product, workflow, customer, and target.",
        "# TYPE wfo_process_count gauge",
        f'wfo_process_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="failed",product_name="Product 2",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 1.0',
        f'wfo_process_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="resumed",product_name="Product 1",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 1.0',
        f'wfo_process_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="completed",product_name="Product 2",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 1.0',
        f'wfo_process_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="suspended",product_name="Product 1",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 1.0',
        f'wfo_process_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="completed",product_name="Product 1",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 1.0',
        f'wfo_process_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="True",last_status="suspended",product_name="Product 1",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 1.0',
        f'wfo_process_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="True",last_status="completed",product_name="Product 2",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 1.0',
        "# HELP wfo_process_seconds_total_count Total time spent on processes in seconds.",
        "# TYPE wfo_process_seconds_total_count gauge",
        f'wfo_process_seconds_total_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="failed",product_name="Product 2",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 600.0',
        f'wfo_process_seconds_total_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="resumed",product_name="Product 1",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 600.0',
        f'wfo_process_seconds_total_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="completed",product_name="Product 2",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 600.0',
        f'wfo_process_seconds_total_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="suspended",product_name="Product 1",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 600.0',
        f'wfo_process_seconds_total_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="False",last_status="completed",product_name="Product 1",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 600.0',
        f'wfo_process_seconds_total_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="True",last_status="suspended",product_name="Product 1",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 600.0',
        f'wfo_process_seconds_total_count{{created_by="None",customer_id="{CUSTOMER_ID}",is_task="True",last_status="completed",product_name="Product 2",workflow_name="workflow_for_testing_processes_py",workflow_target="SYSTEM"}} 600.0',
    ]

    assert all(line in response.text for line in expected_metric_lines)



from pathlib import Path


def absolute_path(path: str) -> str:
    file = Path(__file__).resolve().parent / "data" / path
    return str(file)


def create_main():
    with open("main.py", "w") as fp:
        fp.write(
            "from orchestrator import OrchestratorCore\n"
            "from orchestrator.cli.main import app as core_cli\n"
            "from orchestrator.settings import AppSettings\n"
            "\n"
            "app = OrchestratorCore(base_settings=AppSettings())\n"
            'if __name__ == "__main__":\n'
            "    core_cli()\n"
        )



import pytest
import structlog

from orchestrator.cli.generate import app as generate_app
from test.unit_tests.cli.helpers import absolute_path

logger = structlog.get_logger()


@pytest.mark.parametrize(
    "config_file,expected_exception,expected_message",
    [
        ("invalid_product_config1.yaml", ValueError, "found multiple"),
        ("invalid_product_config2.yaml", ValueError, "Cycle detected"),
    ],
)
def test_product_block_validation(config_file, expected_exception, expected_message, cli_invoke):
    config_file = absolute_path(config_file)
    with pytest.raises(expected_exception, match=expected_message):
        cli_invoke(generate_app, ["product-blocks", "--config-file", config_file])



from uuid import uuid4

import pytest

from orchestrator.db import InputStateTable, ProcessStepTable, ProcessSubscriptionTable, ProcessTable, db
from orchestrator.services.input_state import retrieve_input_state, store_input_state
from orchestrator.workflow import ProcessStatus, StepStatus


@pytest.fixture
def completed_process(test_workflow, generic_subscription_1):
    process_id = uuid4()
    process = ProcessTable(
        process_id=process_id, workflow_id=test_workflow.workflow_id, last_status=ProcessStatus.COMPLETED
    )
    init_step = ProcessStepTable(process_id=process_id, name="Start", status=StepStatus.SUCCESS, state={})
    insert_step = ProcessStepTable(
        process_id=process_id,
        name="Insert UUID in state",
        status=StepStatus.SUCCESS,
        state={"subscription_id": generic_subscription_1},
    )
    check_step = ProcessStepTable(
        process_id=process_id,
        name="Test that it is a string now",
        status=StepStatus.SUCCESS,
        state={"subscription_id": generic_subscription_1},
    )
    step = ProcessStepTable(
        process_id=process_id,
        name="Modify",
        status=StepStatus.SUCCESS,
        state={"subscription_id": generic_subscription_1},
    )

    process_subscription = ProcessSubscriptionTable(process_id=process_id, subscription_id=generic_subscription_1)
    input_state = InputStateTable(process_id=process_id, input_state={"key": "value"}, input_type="initial_state")

    db.session.add(process)
    db.session.add(init_step)
    db.session.add(insert_step)
    db.session.add(check_step)
    db.session.add(step)
    db.session.add(process_subscription)
    db.session.add(input_state)
    db.session.commit()

    return process_id, input_state


def test_retrieve_input_state(completed_process):
    process_id, input_state = completed_process
    retrieved_state = retrieve_input_state(process_id, "initial_state")
    assert retrieved_state == input_state


def test_store_input_state(completed_process):
    process_id, input_state = completed_process
    store_input_state(process_id, input_state.input_state, "user_input")

    states = InputStateTable.query.all()
    assert len(states) == 2



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ipaddress import IPv4Address, IPv6Address
from uuid import UUID

from pydantic import ConfigDict
from structlog import get_logger

from orchestrator.forms.validators import CustomerId
from orchestrator.targets import Target
from orchestrator.types import SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, done, step, workflow
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form
from pydantic_forms.core import FormPage
from pydantic_forms.types import FormGenerator, State, UUIDstr
from test_orchestrator.products.test_product import TestProductInactive

logger = get_logger(__name__)


def initial_input_form_generator(product_name: str, product: UUIDstr) -> FormGenerator:
    class CreateTestProductForm(FormPage):
        model_config = ConfigDict(title=product_name)

        customer_id: CustomerId

        an_int: int
        a_str: str
        a_bool: bool
        an_uuid: UUIDstr
        an_ipv4: IPv4Address
        an_ipv6: IPv6Address

    user_input = yield CreateTestProductForm

    return user_input.model_dump()


@step("Construct Subscription model")
def construct_subscription_model(
    product: UUIDstr,
    customer_id: UUIDstr,
    an_int: int,
    a_str: str,
    a_bool: bool,
    an_uuid: UUIDstr,
    an_ipv4: IPv4Address,
    an_ipv6: IPv6Address,
) -> State:
    test_product = TestProductInactive.from_product_id(
        product_id=product,
        customer_id=customer_id,
        status=SubscriptionLifecycle.INITIAL,
    )
    test_product.testproduct.an_int = an_int
    test_product.testproduct.a_str = a_str
    test_product.testproduct.a_bool = a_bool
    test_product.testproduct.an_uuid = UUID(an_uuid)
    test_product.testproduct.an_ipv4 = an_ipv4
    test_product.testproduct.an_ipv6 = an_ipv6
    test_product.save()
    return {
        "subscription": test_product,
        "subscription_id": test_product.subscription_id,
        "subscription_description": test_product.description,
    }


@workflow("Create a test product", initial_input_form=wrap_create_initial_input_form(initial_input_form_generator))
def create_test_product() -> StepList:
    return begin >> construct_subscription_model >> store_process_subscription(Target.CREATE) >> done



import pytest

from orchestrator.db import SubscriptionTable, db
from orchestrator.targets import Target
from test.unit_tests.workflows import (
    assert_complete,
    extract_state,
    run_workflow,
)


@pytest.mark.workflow
def test_check_subscriptions(generic_subscription_1, validation_workflow_instance):
    product = db.session.get(SubscriptionTable, generic_subscription_1).product
    product.workflows.append(validation_workflow_instance)
    db.session.add(product)
    db.session.commit()

    init_data = {
        "product_type": "Generic",
    }

    result, process, step_log = run_workflow("task_validate_product_type", init_data)

    assert_complete(result)

    state = extract_state(result)

    assert state["product_type"] == "Generic"
    assert state["workflow_name"] == "task_validate_product_type"
    assert state["workflow_target"] == Target.SYSTEM

    result = state["result"]

    assert len(result) == 1
    assert result[0]["total_workflows_validated"] == 1
    assert len(result[0]["workflows"]) == 1
    assert result[0]["workflows"][0]["workflow_name"] == "validation_workflow"
    assert result[0]["workflows"][0]["product_type"] == "Generic"



"""add cascade constraint on processes input state.

Revision ID: fc5c993a4b4a
Revises: 42b3d076a85b
Create Date: 2025-04-09 18:27:31.922214

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "fc5c993a4b4a"
down_revision = "42b3d076a85b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Drop the existing foreign key constraint
    op.drop_constraint("input_states_pid_fkey", "input_states", type_="foreignkey")

    # Add a new foreign key constraint with cascade delete
    op.create_foreign_key(
        "input_states_pid_fkey",
        "input_states",
        "processes",
        ["pid"],
        ["pid"],
        ondelete="CASCADE",
    )


def downgrade() -> None:
    # Drop the cascade foreign key constraint
    op.drop_constraint("input_states_pid_fkey", "input_states", type_="foreignkey")

    # Recreate the original foreign key constraint without cascade
    op.create_foreign_key(
        "input_states_pid_fkey",
        "input_states",
        "processes",
        ["pid"],
        ["pid"],
    )



# Copyright 2019-2025 GÉANT, SURF, ESnet
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from orchestrator.utils.state import inject_args
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.utils import ensure_provisioning_status


@step("Dummy Step Success")
def dummy_step_success(state: dict) -> dict:
    """A dummy step that always succeeds."""
    state["dummy_step"] = "success"
    return {"status": "success"}


# Step definition: A test step that always succeeds, wrapped with ensure_provisioning_status
@ensure_provisioning_status
@step("Test Step Success")
def test_step_success(state: dict) -> dict:
    """A test step that always succeeds."""
    state["test_step"] = "success"
    return {"status": "success"}


# Pytest fixture to provide a mock state for testing
@pytest.fixture
def state(generic_subscription_1):
    """Fixture to provide a mock state with a generic subscription."""
    return {"subscription": generic_subscription_1}


# Test case: Ensure provisioning status is correctly applied in the workflow
@inject_args
def test_ensure_provisioning_status_with_decorator(state: dict):
    """Test ensure_provisioning_status with annotation."""
    # Build the workflow with steps
    steps = begin >> test_step_success >> dummy_step_success

    # Verify the workflow structure and sequence of the steps taken
    assert isinstance(steps, StepList)
    assert steps[0].name == "Set subscription to 'provisioning'"
    assert steps[1].name == "Test Step Success"
    assert steps[2].name == "Set subscription to 'active'"
    assert steps[3].name == "Dummy Step Success"

    # Simulate execution of the workflow
    for step_func in steps:
        step_func(state)

    # Assert final state values after workflow execution
    assert state["test_step"] == "success"  # Ensure test step succeeded
    assert state["dummy_step"] == "success"  # Ensure dummy step succeeded



import json
from http import HTTPStatus
from unittest.mock import patch

import pytest

from orchestrator import app_settings
from test.unit_tests.fixtures.workflows import add_soft_deleted_workflows  # noqa: F401


@pytest.fixture(autouse=True)
def _add_soft_deleted_workflows(add_soft_deleted_workflows):  # noqa: F811
    add_soft_deleted_workflows(10)


@pytest.fixture
def seed_workflows():
    pass


def get_workflows_query(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = """
query WorkflowsQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!], $query: String) {
  workflows(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy, query: $query) {
    page {
      workflowId
      name
      description
      createdAt
      steps {
        name
        assignee
      }
      products {
        name
        description
        tag
      }
    }
    pageInfo {
      endCursor
      hasNextPage
      hasPreviousPage
      startCursor
      totalItems
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "WorkflowsQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def test_workflows_query(test_client):
    data = get_workflows_query(first=2)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    workflows_data = result["data"]["workflows"]
    workflows = workflows_data["page"]
    pageinfo = workflows_data["pageInfo"]

    assert "errors" not in result
    assert len(workflows) == 2

    assert all(len(workflow["steps"]) > 0 for workflow in workflows)

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 5,
    }


def test_workflows_has_previous_page(test_client):
    data = get_workflows_query(after=1, sort_by=[{"field": "name", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    workflows_data = result["data"]["workflows"]
    workflows = workflows_data["page"]
    pageinfo = workflows_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasNextPage": False,
        "hasPreviousPage": True,
        "startCursor": 1,
        "endCursor": 4,
        "totalItems": 5,
    }

    assert len(workflows) == 4
    assert [workflow["name"] for workflow in workflows] == [
        "task_clean_up_tasks",
        "task_resume_workflows",
        "task_validate_product_type",
        "task_validate_products",
    ]


@pytest.mark.parametrize(
    "query_args",
    [
        {"filter_by": [{"field": "name", "value": "task_"}]},
        {"query_string": "name:task_*"},
    ],
)
def test_workflows_filter_by_name(test_client, query_args):
    with patch.object(app_settings, "FILTER_BY_MODE", "partial"):
        data = get_workflows_query(**query_args, sort_by=[{"field": "name", "order": "ASC"}])
        response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    workflows_data = result["data"]["workflows"]
    workflows = workflows_data["page"]
    pageinfo = workflows_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "endCursor": 3,
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "totalItems": 4,
    }
    expected_workflows = [
        "task_clean_up_tasks",
        "task_resume_workflows",
        "task_validate_product_type",
        "task_validate_products",
    ]
    assert [rt["name"] for rt in workflows] == expected_workflows


@pytest.mark.parametrize(
    "query_args",
    [
        {"filter_by": [{"field": "product", "value": "Product 1"}]},
        {"query_string": 'product:"Product 1"'},
    ],
)
def test_workflows_filter_by_product(test_client, query_args):
    data = get_workflows_query(**query_args, sort_by=[{"field": "name", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    workflows_data = result["data"]["workflows"]
    workflows = workflows_data["page"]
    pageinfo = workflows_data["pageInfo"]
    assert "errors" not in result
    assert pageinfo == {
        "endCursor": 0,
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "totalItems": 1,
    }
    assert workflows[0]["name"] == "modify_note"


@pytest.mark.parametrize(
    "query_args",
    [{"query_string": "tag:gen1"}, {"query_string": "tag:gen2"}, {"query_string": "product:Product"}],
)
def test_workflows_filter_by_tag(test_client, query_args):
    with patch.object(app_settings, "FILTER_BY_MODE", "partial"):
        data = get_workflows_query(**query_args, sort_by=[{"field": "name", "order": "ASC"}])
        response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    workflows_data = result["data"]["workflows"]
    workflows = workflows_data["page"]
    pageinfo = workflows_data["pageInfo"]
    assert "errors" not in result
    assert pageinfo == {
        "endCursor": 0,
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "totalItems": 1,
    }
    assert len(workflows) == 1
    assert workflows[0]["name"] == "modify_note"


def test_workflows_sort_by_resource_type_desc(test_client):
    data = get_workflows_query(sort_by=[{"field": "name", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()

    workflows_data = result["data"]["workflows"]
    workflows = workflows_data["page"]
    pageinfo = workflows_data["pageInfo"]

    assert pageinfo == {
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "endCursor": 4,
        "totalItems": 5,
    }
    expected_workflows = [
        "task_validate_products",
        "task_validate_product_type",
        "task_resume_workflows",
        "task_clean_up_tasks",
        "modify_note",
    ]
    assert [rt["name"] for rt in workflows] == expected_workflows



from contextlib import contextmanager

from oauth2_lib.settings import oauth2lib_settings


@contextmanager
def mutation_authorization():
    old_oauth2_active = oauth2lib_settings.OAUTH2_ACTIVE
    old_mutations_enabled = oauth2lib_settings.MUTATIONS_ENABLED

    oauth2lib_settings.OAUTH2_ACTIVE = False
    oauth2lib_settings.MUTATIONS_ENABLED = True

    yield

    oauth2lib_settings.OAUTH2_ACTIVE = old_oauth2_active
    oauth2lib_settings.MUTATIONS_ENABLED = old_mutations_enabled



import asyncio
from http import HTTPStatus
from threading import Event
from time import sleep
from unittest import mock
from uuid import uuid4

import pytest
from pydantic_i18n import PydanticI18n
from sqlalchemy import select

from orchestrator.config.assignee import Assignee
from orchestrator.db import ProcessStepTable, ProcessTable, db
from orchestrator.services.processes import (
    SYSTEM_USER,
    _async_resume_processes,
    _db_create_process,
    _db_log_process_ex,
    _db_log_step,
    _run_process_async,
    load_process,
    safe_logstep,
    start_process,
)
from orchestrator.services.settings import get_engine_settings
from orchestrator.settings import app_settings
from orchestrator.targets import Target
from orchestrator.utils.errors import ApiException, error_state_to_dict
from orchestrator.workflow import (
    Abort,
    Complete,
    Failed,
    ProcessStat,
    ProcessStatus,
    Skipped,
    StepList,
    StepStatus,
    Success,
    Suspend,
    Waiting,
    done,
    init,
    make_step_function,
    make_workflow,
    step,
    workflow,
)
from pydantic_forms.core.translations import translations
from pydantic_forms.exceptions import FormValidationError
from test.unit_tests.workflows import WorkflowInstanceForTests, run_workflow, store_workflow


def _workflow_test_fn():
    pass


def _get_process_steps(process_id, *, order_by=None):
    stmt = select(ProcessStepTable).filter_by(process_id=process_id)
    if order_by is not None:
        stmt = stmt.order_by(order_by)
    return db.session.scalars(stmt).all()


@pytest.fixture
def simple_workflow():
    wf = make_workflow(_workflow_test_fn, "wf description", None, Target.SYSTEM, StepList())
    wf.name = "wf name"
    return store_workflow(wf)


def test_db_create_process(simple_workflow):
    process_id = uuid4()
    pstat = ProcessStat(process_id, simple_workflow, None, None, current_user="user")

    _db_create_process(pstat)

    process = db.session.get(ProcessTable, process_id)
    assert process
    assert process.workflow.name == "wf name"
    assert process.is_task


def test_process_log_db_step_success(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Success(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "success"
    assert psteps[0].process_id == process_id
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.RUNNING
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 2
    assert psteps[1].status == "success"
    assert psteps[1].process_id == process_id
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.RUNNING
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_skipped(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Skipped(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "skipped"
    assert psteps[0].process_id == process_id
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.RUNNING
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 2
    assert psteps[1].status == "skipped"
    assert psteps[1].process_id == process_id
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.RUNNING
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_suspend(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Suspend(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "suspend"
    assert psteps[0].process_id == process_id
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.SUSPENDED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 2
    assert psteps[1].status == "suspend"
    assert psteps[1].process_id == process_id
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.SUSPENDED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_waiting(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = "expected failure"
    state = Waiting(error_state_to_dict(Exception(state_data)))

    result = _db_log_step(pstat, step, state)
    assert result.iswaiting()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "waiting"
    assert psteps[0].process_id == process_id
    assert psteps[0].state["error"] == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.WAITING
    assert p.last_step == "step"
    assert p.failed_reason == state_data
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert result.iswaiting()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "waiting"
    assert psteps[0].process_id == process_id
    pstep_state = psteps[0].state
    assert len(pstep_state["executed_at"]) == 1
    del pstep_state["executed_at"]
    assert pstep_state == {"class": "Exception", "error": state_data, "retries": 1}
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.WAITING
    assert p.last_step == "step"
    assert p.failed_reason == state_data
    assert p.assignee == "assignee"


def test_process_log_db_step_failed(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, Assignee.SYSTEM)

    state_data = Exception("Hard failure")
    state = Failed(state_data)

    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].process_id == process_id
    saved_state = error_state_to_dict(state_data)
    saved_state.pop("traceback")
    assert psteps[0].state == saved_state
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.FAILED
    assert p.last_step == "step"
    assert p.failed_reason == "Hard failure"
    assert p.assignee == Assignee.SYSTEM

    step = make_step_function(lambda: None, "step", None, Assignee.SYSTEM)
    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].process_id == process_id
    pstep_state = psteps[0].state
    assert len(pstep_state["executed_at"]) == 1
    del pstep_state["executed_at"]
    del pstep_state["class"]
    assert pstep_state == {"error": "Hard failure", "retries": 1}
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.FAILED
    assert p.last_step == "step"
    assert p.failed_reason == "Hard failure"
    assert p.assignee == Assignee.SYSTEM


def test_process_log_db_step_assertion_failed(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
        is_task=True,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")

    state_data = AssertionError("Assertion failure")
    state = Failed(state_data)

    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].process_id == process_id
    saved_state = error_state_to_dict(state_data)
    saved_state.pop("traceback")
    assert psteps[0].state == saved_state
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.INCONSISTENT_DATA
    assert p.last_step == "step"
    assert p.failed_reason == "Assertion failure"
    assert p.assignee == Assignee.NOC

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].process_id == process_id
    pstep_state = psteps[0].state
    assert len(pstep_state["executed_at"]) == 1
    del pstep_state["executed_at"]
    del pstep_state["class"]
    assert pstep_state == {"error": "Assertion failure", "retries": 1}
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.INCONSISTENT_DATA
    assert p.last_step == "step"
    assert p.failed_reason == "Assertion failure"
    assert p.assignee == Assignee.NOC


def test_process_log_db_step_api_failed(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
        is_task=True,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")

    state_data = ApiException(HTTPStatus.BAD_GATEWAY, "API failure")
    state = Failed(state_data)

    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].process_id == process_id
    saved_state = error_state_to_dict(state_data)
    saved_state.pop("traceback")
    assert psteps[0].state == saved_state
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.API_UNAVAILABLE
    assert p.last_step == "step"
    assert p.failed_reason == "API failure"
    assert p.assignee == Assignee.SYSTEM

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].process_id == process_id
    pstep_state = psteps[0].state
    assert len(pstep_state["executed_at"]) == 1
    del pstep_state["executed_at"]
    del pstep_state["class"]
    assert pstep_state == {
        "error": "API failure",
        "retries": 1,
        "headers": "",
        "status_code": HTTPStatus.BAD_GATEWAY,
        "body": None,
    }
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.API_UNAVAILABLE
    assert p.last_step == "step"
    assert p.failed_reason == "API failure"
    assert p.assignee == Assignee.SYSTEM


def test_process_log_db_step_abort(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Abort(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "abort"
    assert psteps[0].process_id == process_id
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.ABORTED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 2
    assert psteps[1].status == "abort"
    assert psteps[1].process_id == process_id
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.ABORTED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_complete(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Complete(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 1
    assert psteps[0].status == "complete"
    assert psteps[0].process_id == process_id
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.COMPLETED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = _get_process_steps(p.process_id)

    assert len(psteps) == 2
    assert psteps[1].status == "complete"
    assert psteps[1].process_id == process_id
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.COMPLETED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_no_process_id(simple_workflow):
    process_id = uuid4()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Complete(state_data)

    with pytest.raises(ValueError) as exc_info:
        _db_log_step(pstat, step, state)
    assert f"Failed to write failure step to process: process with PID {process_id} not found" in str(exc_info.value)


def test_process_log_db_step_deduplication(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")

    step1 = make_step_function(lambda: None, "step1", None, Assignee.SYSTEM)
    step2 = make_step_function(lambda: None, "step2", None, Assignee.SYSTEM)
    error_state_data = {"error": "Soft failure"}
    _db_log_step(pstat, step1, Failed(error_state_data))
    _db_log_step(pstat, step2, Failed(error_state_data))
    _db_log_step(pstat, step1, Failed(error_state_data))
    _db_log_step(pstat, step1, Success({}))
    _db_log_step(pstat, step1, Failed(error_state_data))

    psteps = _get_process_steps(p.process_id, order_by=ProcessStepTable.executed_at.asc())

    assert psteps[0].name == "step1"
    assert psteps[0].status == "failed"
    assert psteps[0].state == error_state_data
    assert psteps[0].created_by == "user"

    assert psteps[1].name == "step2"
    assert psteps[1].status == "failed"
    assert psteps[1].state == error_state_data
    assert psteps[1].created_by == "user"

    assert psteps[2].name == "step1"
    assert psteps[2].status == "failed"
    assert psteps[2].state == error_state_data
    assert psteps[2].created_by == "user"

    assert psteps[3].name == "step1"
    assert psteps[3].status == "success"
    assert psteps[3].state == {}
    assert psteps[3].created_by == "user"

    assert psteps[4].name == "step1"
    assert psteps[4].status == "failed"
    assert psteps[4].state == error_state_data
    assert psteps[4].created_by == "user"

    assert len(psteps) == 5

    assert p.last_status == ProcessStatus.FAILED
    assert p.last_step == "step1"
    assert p.failed_reason == "Soft failure"
    assert p.assignee == Assignee.SYSTEM


def test_safe_logstep(simple_workflow):
    process_id = uuid4()
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"not_serializable": object()}
    state = Complete(state_data)

    with mock.patch("orchestrator.services.processes._db_log_step", spec=_db_log_step) as mock__db_log_step:
        mock__db_log_step.side_effect = [
            Exception("Failed to commit because of json serializable failure"),
            mock.sentinel.result,
        ]
        result = safe_logstep(pstat, step, state)

        assert result == mock.sentinel.result
        mock__db_log_step.assert_has_calls(
            [
                mock.call(
                    pstat,
                    step,
                    state,
                    None,  # broadcast_func, which is None for in-memory broadcasting
                ),
                mock.call(
                    pstat,
                    step,
                    Failed(
                        {
                            "class": "Exception",
                            "error": "Failed to commit because of json serializable failure",
                            "traceback": mock.ANY,
                        }
                    ),
                    None,  # broadcast_func, which is None for in-memory broadcasting
                ),
            ]
        )


def test_safe_logstep_critical_failure():
    process_id = uuid4()

    # Skip storing the actual process to let `_db_log_step` fail

    pstat = ProcessStat(process_id, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"not_serializable": object()}
    state = Complete(state_data)

    with mock.patch(  # Mock just to be able to spy on calls
        "orchestrator.services.processes._db_log_step", spec=_db_log_step, side_effect=_db_log_step
    ) as mock__db_log_step:
        with pytest.raises(ValueError) as e:
            result = safe_logstep(pstat, step, state)

            assert result == mock.sentinel.result

            mock__db_log_step.assert_has_calls(
                [
                    mock.call(pstat, step, state),
                    mock.call(
                        pstat,
                        step,
                        Failed(
                            {
                                "class": "Exception",
                                "error": f"Failed to write failure step to process: process with PID {process_id} not found",
                                "traceback": mock.ANY,
                            }
                        ),
                    ),
                ]
            )

        assert f"Failed to write failure step to process: process with PID {process_id} not found" in str(e.value)


@mock.patch("orchestrator.services.processes._get_process")
@mock.patch("orchestrator.services.processes.resume_process")
def test_async_resume_processes(mock_resume_process, mock_get_process, caplog):
    """Test that _async_resume_process() rejects running processes and handles failures."""
    processes = [
        mock.Mock(spec=ProcessTable()),
        mock.Mock(spec=ProcessTable()),
        mock.Mock(spec=ProcessTable()),
        mock.Mock(spec=ProcessTable()),
    ]
    processes[0].process_id, processes[0].last_status = 123, ProcessStatus.RUNNING
    processes[1].process_id, processes[1].last_status = 124, ProcessStatus.FAILED
    processes[2].process_id, processes[2].last_status = 125, ProcessStatus.API_UNAVAILABLE
    processes[3].process_id, processes[3].last_status = 126, ProcessStatus.RESUMED

    # get_process() should be called 4 times
    mock_get_process.side_effect = processes

    # resume_process() should be called 2 times for the non-running / non-resumed processes; let 1 call fail
    mock_resume_process.side_effect = [None, ValueError("This workflow cannot be resumed")]

    # Don't set app_settings.TESTING=False because we want to await the result
    asyncio.run(_async_resume_processes(processes, "testusername"))

    assert len(mock_get_process.mock_calls) == 4
    assert len(mock_resume_process.mock_calls) == 2
    assert "Cannot resume a running process" in caplog.text  # process_id 123 should not be resumed
    assert "Cannot resume a resumed process" in caplog.text  # process_id 126 should not be resumed
    assert "Failed to resume process" in caplog.text  # process_id 125 should fail
    assert "Completed resuming processes" in caplog.text


def test_db_log_process_ex(simple_workflow):
    process_id = uuid4()

    # No exceptions!
    _db_log_process_ex(process_id, ValueError())

    # Now with existing process
    p = ProcessTable(
        process_id=process_id,
        workflow_id=simple_workflow.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=SYSTEM_USER,
    )
    db.session.add(p)
    db.session.commit()

    _db_log_process_ex(process_id, ValueError())

    assert p.last_status == "failed"


def test_run_process_async_success():
    process_id = uuid4()
    event = Event()

    def run_func():
        return Success(event.wait(1))

    # Disable Testing setting since we want to run async
    app_settings.TESTING = False
    _run_process_async(process_id, run_func)
    sleep(0.01)

    assert get_engine_settings().running_processes == 1

    event.set()
    sleep(1)

    assert get_engine_settings().running_processes == 0
    app_settings.TESTING = True


@mock.patch("orchestrator.services.processes._db_log_process_ex")
def test_run_process_async_exception(mock_db_log_process_ex):
    process_id = uuid4()
    event = Event()

    def run_func():
        event.wait(1)
        raise ValueError("Failed")

    # Disable Testing setting since we want to run async
    app_settings.TESTING = False
    _run_process_async(process_id, run_func)
    sleep(0.1)

    assert get_engine_settings().running_processes == 1

    event.set()
    sleep(0.1)

    assert get_engine_settings().running_processes == 0

    mock_db_log_process_ex.assert_called_once_with(process_id, mock.ANY)
    assert repr(mock_db_log_process_ex.call_args[0][1]) == "ValueError('Failed')"
    app_settings.TESTING = True


@mock.patch("orchestrator.services.processes.store_input_state")
@mock.patch("orchestrator.services.processes._run_process_async", return_value=(mock.sentinel.process_id))
@mock.patch("orchestrator.services.processes._db_create_process")
@mock.patch("orchestrator.services.processes.post_form")
@mock.patch("orchestrator.services.processes.get_workflow")
def test_start_process(
    mock_get_workflow, mock_post_form, mock_db_create_process, mock_run_process_async, mock_store_input_state
):
    @step("test step")
    def test_step():
        pass

    wf = make_workflow(lambda: None, "description", None, Target.SYSTEM, [test_step])
    wf.name = "name"
    mock_get_workflow.return_value = wf

    mock_post_form.return_value = {"a": 1}

    result = start_process(mock.sentinel.wf_name, [{"a": 2}], mock.sentinel.user)

    initial_state = {
        "a": 1,
        "process_id": mock.ANY,
        "reporter": mock.sentinel.user,
        "workflow_name": mock.sentinel.wf_name,
        "workflow_target": Target.SYSTEM,
    }
    mock_store_input_state.assert_called_once_with(mock.ANY, initial_state, "initial_state")
    pstat = mock_db_create_process.call_args[0][0]
    assert result == mock.sentinel.process_id
    assert pstat.current_user == mock.sentinel.user
    assert pstat.state.status == StepStatus.SUCCESS
    assert pstat.workflow == wf
    assert pstat.state.unwrap() == {
        "process_id": mock.ANY,
        "reporter": mock.sentinel.user,
        "workflow_name": mock.sentinel.wf_name,
        "workflow_target": Target.SYSTEM,
        "a": 1,
    }
    assert pstat.log == [test_step]
    mock_post_form.assert_called_once_with(
        mock.ANY,
        {
            "process_id": mock.ANY,
            "reporter": mock.sentinel.user,
            "workflow_name": mock.sentinel.wf_name,
            "workflow_target": Target.SYSTEM,
        },
        [{"a": 2}],
    )
    mock_get_workflow.assert_called_once_with(mock.sentinel.wf_name)

    mock_post_form.reset_mock()

    class MockEmptyValidationError:
        def errors(self):
            return []

    tr = PydanticI18n(translations)
    mock_post_form.side_effect = FormValidationError("", MockEmptyValidationError(), tr=tr)

    with pytest.raises(FormValidationError):
        start_process(mock.sentinel.wf_name, None, mock.sentinel.user)
    mock_post_form.assert_called_once_with(
        mock.ANY,
        {
            "process_id": mock.ANY,
            "reporter": mock.sentinel.user,
            "workflow_name": mock.sentinel.wf_name,
            "workflow_target": Target.SYSTEM,
        },
        [{}],
    )


@step("Step 1")
def step1():
    return {"value": 1}


@step("Step 2")
def step2():
    return {"value": 2}


def get_step_names(process):
    return [fn.name for fn in process.log]


@pytest.mark.parametrize(
    "num_steps_finished,step_names",
    [
        (0, (["Start", "Step 1", "Done"], ["Start", "Step 1", "Step 2", "Done"], ["Start", "Done"])),
        (1, (["Step 1", "Done"], ["Step 1", "Step 2", "Done"], ["Done"])),
        (2, (["Done"], ["Step 2", "Done"], [])),  # WF with step removed is missing last step
        (3, ([], [], [])),  # WF is complete
    ],
)
def test_load_process_with_altered_steps(num_steps_finished, step_names):

    # Run original workflow with 3 steps
    with WorkflowInstanceForTests(workflow("Test wf")(lambda: init >> step1 >> done), "test_wf"):
        _, p_stat, steps = run_workflow("test_wf", [{}])
        process_table = db.session.get(ProcessTable, p_stat.process_id)

        for step_fn, wf_process in steps[:num_steps_finished]:
            process_table.steps.append(
                ProcessStepTable(
                    process_id=p_stat.process_id, name=step_fn.name, status=wf_process.status, state=wf_process.unwrap()
                )
            )

        process = load_process(process_table)
        assert get_step_names(process) == step_names[0]

    # Load process for workflow with step added at the end
    with WorkflowInstanceForTests(workflow("Test wf")(lambda: init >> step1 >> step2 >> done), "test_wf"):
        process = load_process(process_table)
        assert get_step_names(process) == step_names[1]

    # Load process for workflow with step removed at the end
    with WorkflowInstanceForTests(workflow("Test wf")(lambda: init >> done), "test_wf"):
        process = load_process(process_table)
        assert get_step_names(process) == step_names[2]



import json
from unittest.mock import patch

import pytest
from sqlalchemy import select, text

from orchestrator.cli.database import migrate_domain_models
from orchestrator.db import db
from orchestrator.db.models import ResourceTypeTable, SubscriptionInstanceValueTable
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import ProductBlockModel, SubscriptionModel
from orchestrator.services.resource_types import get_resource_types
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum


def test_migrate_domain_models_new_product_block(
    test_product_type_one, test_product_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert "new_block" not in subscription.block.model_dump()

    # The subscription should be what we expect before altering the registries/database
    assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class TestBlock(ProductBlockModel, product_block_name="test block", lifecycle=[SubscriptionLifecycle.ACTIVE]):
            str_field: str

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            new_block: TestBlock
            block: ProductBlockOneForTest

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        inputs = json.dumps(
            {
                "test block": {"description": "test block description", "tag": "test_block_tag"},
                "str_field": {"test block": "test"},
            }
        )
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

        assert len(upgrade_sql) == 5
        assert len(downgrade_sql) == 5

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        new_model: ProductTypeOneForTestNew = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert new_model.new_block
        assert new_model.new_block.str_field
        assert new_model.new_block.str_field == "test"

        # Restore database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()


@pytest.mark.xfail(reason="#888 generator bug - domain_model_attr missing on subscription_instance_relations inserts")
def test_migrate_domain_models_new_product_block_on_product_block(
    test_product_type_one, test_product_block_one, product_one_subscription_1, test_product_sub_block_one
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert "new_block" not in subscription.block.model_dump()

    # The subscription should be what we expect before altering the registries/database
    assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class TestBlock(ProductBlockModel, product_block_name="test block", lifecycle=[SubscriptionLifecycle.ACTIVE]):
            str_field: str

        class ProductBlockOneForTestUpdated(
            ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            sub_block: SubBlockOneForTest
            sub_block_2: SubBlockOneForTest
            sub_block_list: list[SubBlockOneForTest]
            str_field: str
            int_field: int
            list_field: list[int]
            enum_field: DummyEnum
            new_block: TestBlock  # <- new product block field

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            block: ProductBlockOneForTestUpdated

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        inputs = json.dumps(
            {
                "test block": {"description": "test block description", "tag": "test_block_tag"},
                "str_field": {"test block": "test"},
            }
        )
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

        assert len(upgrade_sql) == 6
        assert len(downgrade_sql) == 5

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        # TODO (#888)  upgrade_sql contains insert statements for subscription_instance_relations
        #  without `domain_model_attr` set. Therefore we cannot reliably map instances of TestBlock to the
        #  `new_block` field. SubscriptionModel._load_instances() would "fallback" to the field name in the python model
        #  and when reconstructing instances (which has its flaws as well).
        #  SubscriptionModel._load_root_instance() can't do this as the instance data is aggregated in the DB query.
        new_model: ProductTypeOneForTestNew = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert new_model.block.new_block
        assert new_model.block.new_block.str_field
        assert new_model.block.new_block.str_field == "test"

        # Restore database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()


def test_migrate_domain_models_new_resource_type(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert "new_int_field" not in subscription.block.model_dump()

        new_int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "new_int_field"])
        assert not new_int_field_resource

    # The subscription should be what we expect before altering the registries/database
    assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class ProductBlockOneForTestUpdated(
            ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            sub_block: SubBlockOneForTest
            sub_block_2: SubBlockOneForTest
            sub_block_list: list[SubBlockOneForTest]
            int_field: int
            str_field: str
            list_field: list[int]
            enum_field: DummyEnum
            new_int_field: int

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            block: ProductBlockOneForTestUpdated

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        inputs = json.dumps(
            {"new_int_field": {"ProductBlockOneForTest": "1", "description": "test new int field type"}}
        )
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

        assert len(upgrade_sql) == 3
        assert len(downgrade_sql) == 4

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        updated_subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert updated_subscription.block
        assert updated_subscription.block.new_int_field == 1

        # Restore database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()


def test_migrate_domain_models_new_product_block_and_resource_type(
    test_product_type_one, test_product_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert "new_block" not in subscription.block.model_dump()

    # The subscription should be what we expect before altering the registries/database
    assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class TestBlock(ProductBlockModel, product_block_name="test block", lifecycle=[SubscriptionLifecycle.ACTIVE]):
            str_field: str
            new_str_field: str

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            new_block: TestBlock
            block: ProductBlockOneForTest

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        inputs = json.dumps(
            {
                "test block": {"description": "test block description", "tag": "test_block_tag"},
                "str_field": {"test block": "test"},
                "new_str_field": {"description": "new str field desc", "test block": "new test"},
            }
        )
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

        assert len(upgrade_sql) == 8
        assert len(downgrade_sql) == 9

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        new_model: ProductTypeOneForTestNew = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert new_model.new_block
        assert new_model.new_block.str_field
        assert new_model.new_block.str_field == "test"
        assert new_model.new_block.new_str_field == "new test"

        # Restore database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()


def test_migrate_domain_models_update_resource_type(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.list_field == [10, 20, 30]
        assert "new_int_field" not in subscription.block.model_dump()

        new_int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "new_int_field"])
        assert not new_int_field_resource

    # The subscription should be what we expect before altering the registries/database
    assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class ProductBlockOneForTestUpdated(
            ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            sub_block: SubBlockOneForTest
            sub_block_2: SubBlockOneForTest
            sub_block_list: list[SubBlockOneForTest]
            str_field: str
            int_field: int
            enum_field: DummyEnum
            new_list_field: list[int]

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            block: ProductBlockOneForTestUpdated

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        updates = json.dumps({"resource_types": {"list_field": "new_list_field"}})
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, updates=updates)

        assert len(upgrade_sql) == 1
        assert "UPDATE" in upgrade_sql[0]
        assert len(downgrade_sql) == 1
        assert "UPDATE" in downgrade_sql[0]

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        updated_subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert updated_subscription.block.new_list_field == [10, 20, 30]

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()


def test_migrate_domain_models_create_and_rename_resource_type(test_product_type_one, product_one_subscription_1):
    _, _, ProductTypeOneForTest = test_product_type_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.list_field == [10, 20, 30]
        assert "new_int_field" not in subscription.block.model_dump()
        assert "new_int_field" not in subscription.block.sub_block.model_dump()

        new_int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "new_int_field"])
        assert not new_int_field_resource

    # The subscription should be what we expect before altering the registries/database
    assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class SubBlockOneForTestNewResource(
            ProductBlockModel, product_block_name="SubBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            int_field: int
            str_field: str
            new_list_field: list[int]

        class ProductBlockOneForTestUpdated(
            ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            sub_block: SubBlockOneForTestNewResource
            sub_block_2: SubBlockOneForTestNewResource
            sub_block_list: list[SubBlockOneForTestNewResource]
            str_field: str
            int_field: int
            enum_field: DummyEnum
            new_list_field: list[int]

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            block: ProductBlockOneForTestUpdated

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        inputs = json.dumps({"new_list_field": {"SubBlockOneForTest": "5"}})
        updates = json.dumps({"resource_types": {"list_field": "new_list_field"}})
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

        assert len(upgrade_sql) == 3
        assert [sql_stmt for sql_stmt in upgrade_sql if "UPDATE" in sql_stmt]
        assert len(downgrade_sql) == 3
        assert [sql_stmt for sql_stmt in downgrade_sql if "UPDATE" in sql_stmt]

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        updated_instance = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert updated_instance.block.new_list_field == [10, 20, 30]
        assert updated_instance.block.sub_block.new_list_field == [5]

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()


def test_migrate_domain_models_create_and_rename_and_delete_resource_type(
    test_product_type_sub_list_union,
    product_sub_list_union_subscription_1,
):
    _, _, ProductSubListUnion = test_product_type_sub_list_union

    def assert_subscription_has_initial_values():
        subscription = ProductSubListUnion.from_subscription(product_sub_list_union_subscription_1)
        assert subscription.test_block.int_field == 1
        assert subscription.test_block.list_union_blocks[0].int_field == 1
        assert "int_field" not in subscription.test_block.list_union_blocks[1].model_dump()
        assert "changed_int_field" not in subscription.test_block.model_dump()
        assert "changed_int_field" not in subscription.test_block.list_union_blocks[0].model_dump()
        assert "changed_int_field" not in subscription.test_block.list_union_blocks[1].model_dump()

        new_int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "new_int_field"])
        assert not new_int_field_resource

    # The subscription should be what we expect before altering the registries/database
    assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class SubBlockOneForTestChanged(
            ProductBlockModel, product_block_name="SubBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            changed_int_field: int

        class SubBlockTwoForTestChanged(
            ProductBlockModel, product_block_name="SubBlockTwoForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            int_field_2: int
            changed_int_field: int

        class ProductBlockWithListUnionForTestNew(
            ProductBlockModel,
            product_block_name="ProductBlockWithListUnionForTest",
            lifecycle=[SubscriptionLifecycle.ACTIVE],
        ):
            list_union_blocks: list[SubBlockTwoForTestChanged | SubBlockOneForTestChanged]
            changed_int_field: int  # <- resource type field renamed from `int_field`
            str_field: str
            list_field: list[int]

        class ProductSubListUnionTest(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_block: ProductBlockWithListUnionForTestNew

        SUBSCRIPTION_MODEL_REGISTRY["ProductSubListUnion"] = ProductSubListUnionTest

        inputs = json.dumps({"changed_int_field": {"SubBlockOneForTest": "5", "SubBlockTwoForTest": "2"}})
        updates = json.dumps({"resource_types": {"int_field": "changed_int_field"}})
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

        assert len(upgrade_sql) == 5
        assert [sql_stmt for sql_stmt in upgrade_sql if "UPDATE" in sql_stmt]
        assert len(downgrade_sql) == 3
        assert [sql_stmt for sql_stmt in downgrade_sql if "UPDATE" in sql_stmt]

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        updated_subscription = ProductSubListUnionTest.from_subscription(product_sub_list_union_subscription_1)
        assert updated_subscription.test_block.changed_int_field == 1
        assert updated_subscription.test_block.list_union_blocks[0].changed_int_field == 1
        assert updated_subscription.test_block.list_union_blocks[1].changed_int_field == 2
        assert "int_field" not in updated_subscription.test_block.model_dump()
        assert "int_field" not in updated_subscription.test_block.list_union_blocks[0].model_dump()
        assert "int_field" not in updated_subscription.test_block.list_union_blocks[1].model_dump()

        int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "int_field"])
        assert not int_field_resource

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()


def test_migrate_domain_models_remove_product(test_product_type_one, product_one_subscription_1):
    _, _, ProductTypeOneForTest = test_product_type_one

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):
        del SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"]

        inputs = json.dumps({"force_delete_products": "no"})
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

        assert len(upgrade_sql) == 7
        assert len(downgrade_sql) == 0

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()


def test_migrate_domain_models_remove_fixed_input(
    test_product_type_one, test_product_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert not subscription.test_fixed_input

    # The subscription should be what we expect before altering the registries/database
    assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            block: ProductBlockOneForTest

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

        assert len(upgrade_sql) == 1
        assert len(downgrade_sql) == 1

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert "test_fixed_input" not in subscription.model_dump()

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    assert_subscription_has_initial_values()


def test_migrate_domain_models_remove_product_block(test_product_type_one, product_one_subscription_1):
    _, _, ProductTypeOneForTest = test_product_type_one

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

        assert len(upgrade_sql) == 3
        assert len(downgrade_sql) == 0

        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block
        assert subscription.block.int_field

        int_field_resource: list[ResourceTypeTable] = get_resource_types(
            filters=[ResourceTypeTable.resource_type == "int_field"]
        )
        int_field_instance_values = db.session.scalars(
            select(SubscriptionInstanceValueTable).where(
                SubscriptionInstanceValueTable.resource_type_id == int_field_resource[0].resource_type_id
            )
        ).all()
        assert len(int_field_instance_values) == 3

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert "block" not in subscription.model_dump()

        int_field_instance_values = db.session.scalars(
            select(SubscriptionInstanceValueTable).where(
                SubscriptionInstanceValueTable.resource_type_id == int_field_resource[0].resource_type_id
            )
        ).all()
        assert len(int_field_instance_values) == 0

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()


def test_migrate_domain_models_remove_resource_type(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class ProductBlockOneForTest(
            ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            sub_block: SubBlockOneForTest
            sub_block_2: SubBlockOneForTest
            sub_block_list: list[SubBlockOneForTest]
            int_field: int
            str_field: str
            enum_field: DummyEnum

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            block: ProductBlockOneForTest

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

        assert len(upgrade_sql) == 4
        assert len(downgrade_sql) == 0

        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.list_field

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert "list_field" not in subscription.block.model_dump()

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()


def test_migrate_domain_models_update_block_resource_type(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.str_field
        assert "update_str_field" not in subscription.block.model_dump()

        new_str_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "update_str_field"])
        assert not new_str_field_resource
        return subscription.block.str_field

    # The subscription should be what we expect before altering the registries/database
    str_field_value = assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class ProductBlockOneForTestUpdated(
            ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            sub_block: SubBlockOneForTest
            sub_block_2: SubBlockOneForTest
            sub_block_list: list[SubBlockOneForTest]
            int_field: int
            update_str_field: str  # <- resource type field renamed from `str_field`
            list_field: list[int]
            enum_field: DummyEnum

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            block: ProductBlockOneForTestUpdated

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        inputs = json.dumps(
            {
                "update_str_field": {"description": "test"},
                "updates": {"block_resource_types": {"ProductBlockOneForTest": {"str_field": "update_str_field"}}},
            }
        )
        updates = json.dumps({"block_resource_types": {"ProductBlockOneForTest": {"str_field": "update_str_field"}}})
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

        assert len(upgrade_sql) == 3
        assert len(downgrade_sql) == 4

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert subscription.block.update_str_field == str_field_value

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()


def test_migrate_domain_models_rename_and_update_block_resource_type(test_product_type_one, product_one_subscription_1):
    _, _, ProductTypeOneForTest = test_product_type_one

    def assert_subscription_has_initial_values():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.str_field
        assert "update_str_field" not in subscription.block.model_dump()

        new_str_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "update_str_field"])
        assert not new_str_field_resource
        return subscription.block.str_field

    # The subscription should be what we expect before altering the registries/database
    str_field_value = assert_subscription_has_initial_values()

    with patch.dict(ProductBlockModel.registry), patch.dict(SUBSCRIPTION_MODEL_REGISTRY):

        class SubBlockOneForTestUpdated(
            ProductBlockModel, product_block_name="SubBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            int_field: int
            other_str_field: str  # <- resource type field renamed from `str_field`

        class ProductBlockOneForTestUpdated(
            ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
        ):
            sub_block: SubBlockOneForTestUpdated
            sub_block_2: SubBlockOneForTestUpdated
            sub_block_list: list[SubBlockOneForTestUpdated]
            int_field: int
            update_str_field: str  # <- resource type field renamed from `str_field`
            list_field: list[int]
            enum_field: DummyEnum

        class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
            test_fixed_input: bool
            block: ProductBlockOneForTestUpdated

        SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

        inputs = json.dumps({"update_str_field": {"description": "test"}})
        updates = json.dumps(
            {
                "resource_types": {"str_field": "other_str_field"},
                "block_resource_types": {"ProductBlockOneForTest": {"other_str_field": "update_str_field"}},
            }
        )
        upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

        assert len(upgrade_sql) == 6
        assert len(downgrade_sql) == 5

        for stmt in upgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

        subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
        assert subscription.block.update_str_field == str_field_value

        # Revert database to its original state
        for stmt in downgrade_sql:
            db.session.execute(text(stmt))
        db.session.commit()

    # Note that this check is done after patch.dict() restored the registries to their original state.
    # The subscription should now be the same as it was before.
    assert_subscription_has_initial_values()



import json
import pathlib
import tempfile

from orchestrator.services.translations import generate_translations
from orchestrator.settings import app_settings


def test_generate_translations_empty():
    assert generate_translations("") == {}


def test_generate_translations():
    assert generate_translations("en-GB")


def test_generate_translations_custom_file():
    old_value = app_settings.TRANSLATIONS_DIR
    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            with open(f"{temp_dir}/nl-NL.json", "w") as temp_file:
                temp_file.write("[]")

            app_settings.TRANSLATIONS_DIR = pathlib.Path(temp_dir)

            assert generate_translations("nl-NL") == {}

            with open(f"{temp_dir}/nl-NL.json", "w") as temp_file:
                temp_file.write('{"foo": {"key1": "val1", "key2":"val2"}}')

            assert generate_translations("nl-NL") == {"foo": {"key1": "val1", "key2": "val2"}}

            app_settings.TRANSLATIONS_DIR = old_value

    finally:
        app_settings.TRANSLATIONS_DIR = old_value


def test_generate_translations_extend_default():
    old_value = app_settings.TRANSLATIONS_DIR
    custom_translations = {
        "forms": {
            "fields": {"note": "Overwritten note entry", "new_field": "New field entry"},
            "new_form_key": "New form value",
        },
        "new_top_level_key": {"new_key": "new_value"},
        "workflow": {"new_wf_translation": "New workflow entry"},
    }
    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            with open(f"{temp_dir}/en-GB.json", "w") as temp_file:
                temp_file.write(json.dumps(custom_translations))

            app_settings.TRANSLATIONS_DIR = pathlib.Path(temp_dir)
            result = generate_translations("en-GB")

            assert set(result.keys()) == {"forms", "workflow", "new_top_level_key"}
            assert result.get("forms", {}).get("new_form_key") == "New form value"
            assert "modify_note" in result.get("workflow", {})
    finally:
        app_settings.TRANSLATIONS_DIR = old_value



import pytest

from orchestrator.db import WorkflowTable, db
from orchestrator.targets import Target
from orchestrator.utils.datetime import nowtz


@pytest.fixture
def add_soft_deleted_workflows():
    def _add_soft_deleted_workflow(n: int):
        for i in range(n):
            db.session.add(
                WorkflowTable(
                    name=f"deleted_workflow_{i}",
                    description="deleted workflow",
                    target=Target.SYSTEM,
                    deleted_at=nowtz(),
                )
            )
        db.session.commit()

    return _add_soft_deleted_workflow



import json
from unittest import mock

from deepdiff import DeepDiff


def assert_no_diff(expected, actual, exclude_paths=None):
    diff = DeepDiff(expected, actual, ignore_order=True, exclude_paths=exclude_paths)
    prettydiff = f"Difference: {json.dumps(diff, indent=2, default=lambda x: str(x))}"

    assert diff == {}, f"Difference between expected and actual output\n{prettydiff}"


def safe_delete_product_block_id(product_block_class):
    """Safely delete product_block_id from product block class if its defined.

    When a product block is not defined within a fixture function, the product_block_id
    is stored inside the class and is kept through multiple tests,
    which results in a foreign key error product block does not exist.
    """
    try:
        del product_block_class.product_block_id
    except AttributeError:
        pass


# By default Pydantic v2 includes documentation urls in the errors.
# Update these urls when upgrading Pydantic.
URL_MISSING = {"url": mock.ANY}
URL_STR_TYPE = {"url": mock.ANY}
URL_PARSING = {"url": mock.ANY}
URL_VALUE = {"url": mock.ANY}



import pytest
from orchestrator.types import SubscriptionLifecycle

from products.product_types.example4 import Example4
from test.unit_tests.workflows import assert_complete, extract_state, run_workflow


@pytest.mark.workflow
def test_happy_flow(responses, example4_subscription):
    # given

    customer_id = "3f4fc287-0911-e511-80d0-005056956c1a"
    crm = CrmMocks(responses)
    crm.get_customer_by_uuid(customer_id)

    # TODO insert additional mocks, if needed (ImsMocks)

    # when

    init_state = {}

    result, process, step_log = run_workflow(
        "modify_example4",
        [{"subscription_id": example4_subscription}, init_state, {}],
    )

    # then

    assert_complete(result)
    state = extract_state(result)

    example4 = Example4.from_subscription(state["subscription_id"])
    assert example4.status == SubscriptionLifecycle.ACTIVE



import pytest
from pydantic import Field

from orchestrator.db import ProductBlockTable, db
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_block_with_list_union(test_product_sub_block_one, test_product_sub_block_two):
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    SubBlockTwoForTestInactive, SubBlockTwoForTestProvisioning, SubBlockTwoForTest = test_product_sub_block_two

    class ProductBlockWithListUnionForTestInactive(
        ProductBlockModel, product_block_name="ProductBlockWithListUnionForTest"
    ):
        list_union_blocks: list[SubBlockTwoForTestInactive | SubBlockOneForTestInactive]
        int_field: int | None = None
        str_field: str | None = None
        list_field: list[int] = Field(default_factory=list)

    class ProductBlockWithListUnionForTestProvisioning(
        ProductBlockWithListUnionForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        list_union_blocks: list[SubBlockTwoForTestProvisioning | SubBlockOneForTestProvisioning]
        int_field: int
        str_field: str | None = None
        list_field: list[int]

    class ProductBlockWithListUnionForTest(
        ProductBlockWithListUnionForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        list_union_blocks: list[SubBlockTwoForTest | SubBlockOneForTest]
        int_field: int
        str_field: str
        list_field: list[int]

    return (
        ProductBlockWithListUnionForTestInactive,
        ProductBlockWithListUnionForTestProvisioning,
        ProductBlockWithListUnionForTest,
    )


@pytest.fixture
def test_product_block_with_list_union_db(
    test_product_sub_block_one_db,
    test_product_sub_block_two_db,
    resource_type_int,
    resource_type_str,
    resource_type_list,
):
    product_sub_block_one = test_product_sub_block_one_db
    product_sub_block_two = test_product_sub_block_two_db

    product_block_with_list_union = ProductBlockTable(
        name="ProductBlockWithListUnionForTest", description="Test Union Sub Block", tag="TEST", status="active"
    )
    product_block_with_list_union.resource_types = [resource_type_int, resource_type_str, resource_type_list]
    product_block_with_list_union.depends_on = [product_sub_block_one, product_sub_block_two]
    db.session.add(product_block_with_list_union)
    db.session.commit()

    return product_block_with_list_union, product_sub_block_one, product_sub_block_two



from orchestrator.targets import Target
from orchestrator.workflow import (
    DEFAULT_CALLBACK_PROGRESS_KEY,
    DEFAULT_CALLBACK_ROUTE_KEY,
    AwaitingCallback,
    begin,
    callback_step,
    done,
    runwf,
    step,
    workflow,
)
from test.unit_tests.workflows import (
    WorkflowInstanceForTests,
    _store_step,
    assert_awaiting_callback,
    assert_complete,
    assert_state,
    run_workflow,
)


@step("Step 1")
def step1():
    return {"steps": [1]}


@step("Step 2")
def step2(steps):
    return {"steps": [*steps, 2]}


@step("Action")
def action():
    return {"ext_response": "Request received"}


@step("Action - Dry run")
def action_dr():
    return {"phase": "DRY_RUN"}


@step("Action - For real")
def action_fr():
    return {"phase": "FOR_REAL"}


@step("Validate")
def validate_state(phase, callback_result):
    if not callback_result["ext_data"].isnumeric():
        raise ValueError("ext_data must be numeric")
    if phase == "DRY_RUN":
        return {"dr_ext_data": callback_result["ext_data"]}
    if phase == "FOR_REAL":
        return {"ext_data": callback_result["ext_data"]}
    raise AssertionError(f"Unknown phase: {phase}")


@step("Cleanup")
def cleanup():
    return {"phase": None}


cb1 = callback_step("Dry run", action_step=action_dr, validate_step=validate_state)
cb2 = callback_step("For real", action_step=action_fr, validate_step=validate_state)


@workflow("Multiple callback wf", target=Target.CREATE)
def multiple_callback_wf():
    return begin >> cb1 >> cb2 >> cleanup >> done


def test_workflow_with_callback():
    @step("Validate")
    def validate(code):
        if code != "12345":
            raise ValueError("Response code is wrong")
        return {"status": "ok"}

    call_ext_system = callback_step(name="Call ext system", action_step=action, validate_step=validate)

    @workflow("Test wf", target=Target.CREATE)
    def test_wf():
        return begin >> step1 >> call_ext_system >> step2 >> done

    with WorkflowInstanceForTests(test_wf, "test_wf"):
        result, process, step_log = run_workflow("test_wf", {})
        assert_awaiting_callback(result)
        state = result.unwrap()
        assert DEFAULT_CALLBACK_ROUTE_KEY in state
        assert state.get("__sub_step") == "Call ext system - Await callback"

        step_log = [step_log[0]] + [step_log[-1]]

        process = process.update(log=process.workflow.steps[1:], state=AwaitingCallback({**state, "code": "12345"}))
        result = runwf(process, _store_step(step_log))
        assert_complete(result)
        assert_state(result, {"steps": [1, 2], "code": "12345"})


def test_callback_wf_with_custom_callback_route():
    call_ext_system = callback_step(
        name="Call ext system", action_step=action, validate_step=step1, callback_route_key="custom_route_key"
    )

    @workflow("Test wf", target=Target.CREATE)
    def test_wf():
        return begin >> call_ext_system >> done

    with WorkflowInstanceForTests(test_wf, "test_wf"):
        result, process, step_log = run_workflow("test_wf", {})
        assert_awaiting_callback(result)
        state = result.unwrap()
        assert DEFAULT_CALLBACK_ROUTE_KEY not in state
        assert "custom_route_key" in state


def test_wf_with_multiple_callback_steps(test_client):
    with WorkflowInstanceForTests(multiple_callback_wf, "multiple_callback_wf"):
        # Start workflow
        response = test_client.post("/api/processes/multiple_callback_wf", json=[{}])
        assert response.status_code == 201
        process_id = response.json()["id"]

        # Check process status
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        assert response_data["last_status"] == "awaiting_callback"
        state = response_data["current_state"]
        assert state["phase"] == "DRY_RUN"
        assert state["__step_group"] == "Dry run"
        assert state["__sub_step"] == "Dry run - Await callback"
        assert "callback_route" in state

        # Continue workflow 1
        callback_route1 = state["callback_route"]
        response = test_client.post(callback_route1, json={"ext_data": "12345", "other": "useless data"})
        assert response.status_code == 200

        # Check process status
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        assert response_data["last_status"] == "awaiting_callback"
        state = response_data["current_state"]
        assert state["dr_ext_data"] == "12345"
        assert state["phase"] == "FOR_REAL"
        assert state["__step_group"] == "For real"
        assert state["__sub_step"] == "For real - Await callback"
        assert "callback_route" in state

        callback_route2 = state["callback_route"]

        assert callback_route1 != callback_route2, "Randomly generated callback routes should be distinct"

        # Continue workflow 2
        response = test_client.post(callback_route2, json={"ext_data": "56789", "other": "very useful data"})
        assert response.status_code == 200

        # Final check
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()

        assert response_data["last_status"] == "completed"

        state = response_data["current_state"]
        assert state["ext_data"] == "56789"

        assert state["phase"] is None


def test_wf_callback_progress_on_completed_process(test_client):
    # given
    @workflow("Test wf", target=Target.CREATE)
    def test_wf():
        return begin >> step1 >> done

    with WorkflowInstanceForTests(test_wf, "test_wf"):
        response = test_client.post("/api/processes/test_wf", json=[{}])
        assert response.status_code == 201
        process_id = response.json()["id"]

        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        assert response_data["last_status"] == "completed"

        # when
        response = test_client.post(f"api/processes/{process_id}/callback/token123/progress", json={})
        response_data = response.json()

        # then
        assert response.status_code == 409
        assert response_data["detail"] == "This process is not in an awaiting state."


def test_wf_callback_progress_with_incorrect_token(test_client):
    # given
    @workflow("Test wf", target=Target.CREATE)
    def test_wf():
        return begin >> cb1 >> done

    with WorkflowInstanceForTests(test_wf, "test_wf"):
        response = test_client.post("/api/processes/test_wf", json=[{}])
        assert response.status_code == 201
        process_id = response.json()["id"]

        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        assert response_data["last_status"] == "awaiting_callback"

        # when
        response = test_client.post(f"api/processes/{process_id}/callback/incorrectoken/progress", json={})
        response_data = response.json()

        # then
        assert response.status_code == 404
        assert response_data["detail"] == "Invalid token"


def test_wf_callback_progress_with_multiple_callback_steps(test_client):
    with WorkflowInstanceForTests(multiple_callback_wf, "multiple_callback_wf"):
        # Start workflow
        response = test_client.post("/api/processes/multiple_callback_wf", json=[{}])
        assert response.status_code == 201
        process_id = response.json()["id"]

        # Check process status
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        assert response_data["last_status"] == "awaiting_callback"
        state = response_data["current_state"]

        # Update dry step progress with dict - update 1
        callback_route1 = state["callback_route"]
        response = test_client.post(
            f"{callback_route1}/progress",
            json={"update": 1},
        )
        assert response.status_code == 200
        response_data = response.json()

        # Check process status
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        state = response_data["current_state"]
        assert state[DEFAULT_CALLBACK_PROGRESS_KEY] == {"update": 1}

        # Update dry step progress with dict - update 2
        callback_route1 = state["callback_route"]
        response = test_client.post(
            f"{callback_route1}/progress",
            json={"update": 2},
        )
        assert response.status_code == 200
        response_data = response.json()

        # Check process status
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        state = response_data["current_state"]
        assert state[DEFAULT_CALLBACK_PROGRESS_KEY] == {"update": 2}

        # Continue workflow dry step
        callback_route1 = state["callback_route"]
        response = test_client.post(callback_route1, json={"ext_data": "12345"})
        assert response.status_code == 200

        # Check process status
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        assert response_data["last_status"] == "awaiting_callback"
        state = response_data["current_state"]
        assert state["dr_ext_data"] == "12345"
        assert DEFAULT_CALLBACK_PROGRESS_KEY not in state

        # Update real step progress with string - update 1
        callback_route2 = state["callback_route"]
        response = test_client.post(
            f"{callback_route2}/progress", data="update 1", headers={"Content-Type": "text/plain; charset=utf-8"}
        )
        assert response.status_code == 200
        response_data = response.json()

        # Check process status
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        state = response_data["current_state"]
        assert state[DEFAULT_CALLBACK_PROGRESS_KEY] == "update 1"

        # Update real step progress with string - update 2
        callback_route2 = state["callback_route"]
        response = test_client.post(
            f"{callback_route2}/progress", data="update 2", headers={"Content-Type": "text/plain; charset=utf-8"}
        )
        assert response.status_code == 200
        response_data = response.json()

        # Check process status
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        state = response_data["current_state"]
        assert state[DEFAULT_CALLBACK_PROGRESS_KEY] == "update 2"

        # Continue workflow real step
        response = test_client.post(callback_route2, json={"ext_data": "56789"})
        assert response.status_code == 200

        # Final check
        response = test_client.get(f"api/processes/{process_id}")
        response_data = response.json()
        state = response_data["current_state"]

        assert response_data["last_status"] == "completed"

        assert all(
            key in state.keys()
            for key in [
                "phase",
                "ext_data",
                "reporter",
                "process_id",
                "dr_ext_data",
                "workflow_name",
                "callback_route",
                "callback_result",
                "workflow_target",
            ]
        )

        state = response_data["current_state"]
        assert state["ext_data"] == "56789"
        assert DEFAULT_CALLBACK_PROGRESS_KEY not in state



import pytest

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def setup():
    class SubBlockInactive(ProductBlockModel, product_block_name="SubBlock"):
        int_field: int | None = None

    class SubBlockProvisioning(SubBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        int_field: int

    class MainBlockInactive(ProductBlockModel, product_block_name="MainBlock"):
        sub_block1: SubBlockInactive
        sub_block2: SubBlockInactive

    yield SubBlockInactive, SubBlockProvisioning, MainBlockInactive


def test_invalid_lifecycle_status(setup):
    SubBlockInactive, SubBlockProvisioning, MainBlockInactive = setup

    expected_error = (
        r"lifecycle status of the type for the field: "
        r"sub_block1, SubBlockProvisioning \(based on SubBlockInactive\) is not suitable for the "
        r"lifecycle status \(provisioning\) of this model"
    )
    with pytest.raises(AssertionError, match=expected_error):

        class MainBlockProvisioning(MainBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
            sub_block1: SubBlockInactive  # invalid lifecycle
            sub_block2: SubBlockInactive  # invalid lifecycle but currently not mentioned in the error


def test_invalid_lifecycle_status_union(setup):
    SubBlockInactive, SubBlockProvisioning, MainBlockInactive = setup

    expected_error = (
        r"The lifecycle status of the type for the field: "
        r"sub_block2, SubBlockProvisioning \(based on SubBlockInactive\) is not suitable for the "
        r"lifecycle status \(provisioning\) of this model"
    )
    with pytest.raises(AssertionError, match=expected_error):

        class MainBlockProvisioning(MainBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
            sub_block1: SubBlockProvisioning
            sub_block2: SubBlockProvisioning | SubBlockInactive  # first lifecycle is valid, second is not



from unittest.mock import MagicMock

from strawberry.types.nodes import FragmentSpread, SelectedField

from orchestrator.graphql.utils import is_query_detailed

is_test_query_detailed = is_query_detailed(("basic", "name"))


def test_is_test_query_detailed_with_basic_fields():
    info = MagicMock(
        selected_fields=[
            SelectedField(
                name="subscriptions",
                directives={},
                arguments={},
                alias=None,
                selections=[
                    SelectedField(
                        name="page",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="basic", directives={}, arguments={}, alias=None, selections=[]),
                            SelectedField(name="name", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                    SelectedField(
                        name="pageInfo",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="totalItems", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                ],
            ),
        ]
    )

    assert not is_test_query_detailed(info)


def test_is_test_query_detailed_with_basic_gragment_field():
    info = MagicMock(
        selected_fields=[
            SelectedField(
                name="subscriptions",
                directives={},
                arguments={},
                alias=None,
                selections=[
                    SelectedField(
                        name="page",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            FragmentSpread(
                                name="fragment",
                                type_condition="graphql type",
                                directives={},
                                selections=[
                                    SelectedField(name="basic", directives={}, arguments={}, selections=[], alias=None),
                                    SelectedField(name="name", directives={}, arguments={}, selections=[], alias=None),
                                ],
                            )
                        ],
                    ),
                    SelectedField(
                        name="pageInfo",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="totalItems", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                ],
            ),
        ]
    )

    assert not is_test_query_detailed(info)


def test_is_test_query_detailed_with_detailed_fields():
    info = MagicMock(
        selected_fields=[
            SelectedField(
                name="subscriptions",
                directives={},
                arguments={},
                alias=None,
                selections=[
                    SelectedField(
                        name="page",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="detailed", directives={}, arguments={}, alias=None, selections=[]),
                            SelectedField(name="not basic", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                    SelectedField(
                        name="pageInfo",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="totalItems", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                ],
            ),
        ]
    )

    assert is_test_query_detailed(info)


def test_is_test_query_detailed_with_detailed_fragment_field():
    info = MagicMock(
        selected_fields=[
            SelectedField(
                name="subscriptions",
                directives={},
                arguments={},
                alias=None,
                selections=[
                    SelectedField(
                        name="page",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            FragmentSpread(
                                name="fragment",
                                type_condition="graphql type",
                                directives={},
                                selections=[
                                    SelectedField(
                                        name="detailed", directives={}, arguments={}, alias=None, selections=[]
                                    ),
                                    SelectedField(
                                        name="not basic", directives={}, arguments={}, alias=None, selections=[]
                                    ),
                                ],
                            )
                        ],
                    ),
                    SelectedField(
                        name="pageInfo",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="totalItems", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                ],
            ),
        ]
    )

    assert is_test_query_detailed(info)


def test_is_test_query_detailed_with_basic_and_detailed_fields():
    info = MagicMock(
        selected_fields=[
            SelectedField(
                name="subscriptions",
                directives={},
                arguments={},
                alias=None,
                selections=[
                    SelectedField(
                        name="page",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="basic", directives={}, arguments={}, alias=None, selections=[]),
                            SelectedField(name="name", directives={}, arguments={}, alias=None, selections=[]),
                            SelectedField(name="detailed", directives={}, arguments={}, alias=None, selections=[]),
                            SelectedField(name="not basic", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                    SelectedField(
                        name="pageInfo",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="totalItems", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                ],
            ),
        ]
    )

    assert is_test_query_detailed(info)



from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY

from products.product_types.example2 import Example2

SUBSCRIPTION_MODEL_REGISTRY.update(
    {
        "example2": Example2,
        },
)  # fmt:skip
from products.product_types.example1 import Example1

SUBSCRIPTION_MODEL_REGISTRY.update(
    {
        "example1 1": Example1,
        "example1 10": Example1,
        "example1 100": Example1,
        "example1 1000": Example1,
        },
)  # fmt:skip
from products.product_types.example4 import Example4

SUBSCRIPTION_MODEL_REGISTRY.update(
    {
        "example4": Example4,
        },
)  # fmt:skip



"""Testcases for multiple product block fields with the same type."""

import itertools
from uuid import uuid4

import pytest

from orchestrator.db import db
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum


@pytest.fixture
def create_fixtures(test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one):

    ProductTypeOneForTestInactive, _, ProductTypeOneForTest = test_product_type_one
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one

    customer_id = str(uuid4())

    def create_subblock(subscription, value):
        return SubBlockOneForTestInactive.new(subscription.subscription_id, int_field=value, str_field=str(value))

    value_generator = itertools.count(1000)

    def create_subscription(*, int_value, sub_block_values=None, sub_block_list=None, sub_block1=None, sub_block2=None):
        subscription = ProductTypeOneForTestInactive.from_product_id(
            product_id=test_product_model.product_id, customer_id=customer_id, insync=True
        )

        if sub_block_list:
            # If sub_block1/sub_block2 are passed then use them, otherwise create placeholder subscription blocks with arbitrary values
            sub_block_list = sub_block_list
            sub_block1 = sub_block1 or create_subblock(subscription, next(value_generator))
            sub_block2 = sub_block2 or create_subblock(subscription, next(value_generator))
        else:
            # List stays empty, if sub_block1/sub_block2 are passed use them, otherwise create subscription blocks from sub_block_values
            sub_block_list = []
            sub_block1 = sub_block1 or create_subblock(subscription, sub_block_values[0])
            sub_block2 = sub_block2 or create_subblock(subscription, sub_block_values[1])

        subscription.block = ProductBlockOneForTestInactive.new(
            subscription_id=subscription.subscription_id,
            int_field=int_value,
            str_field=str(int_value),
            sub_block=sub_block1,
            sub_block_2=sub_block2,
            sub_block_list=sub_block_list,
            enum_field=DummyEnum.FOO,
        )
        subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.ACTIVE)
        subscription.save()
        db.session.commit()
        return subscription

    yield create_subscription, create_subblock


def test_2_field_blocks_from_current_subscription(test_product_type_one, create_fixtures):
    """Test using blocks of the same type in separate fields when the blocks are owned by the current subscription."""
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, _ = create_fixtures

    subscription_10 = create_subscription(int_value=10, sub_block_values=[100, 101])

    subscription_10 = ProductTypeOneForTest.from_subscription(subscription_10.subscription_id)

    assert subscription_10.block.sub_block.int_field == 100
    assert subscription_10.block.sub_block_2.int_field == 101


def test_2_field_blocks_from_current_and_other_subscription(test_product_type_one, create_fixtures):
    """Test using blocks of the same type in separate fields when one block is owned by the current subscription and another block is owned by a different subscription."""
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, create_subblock = create_fixtures

    subscription_20 = create_subscription(int_value=20, sub_block_values=[200, 201])

    subscription_10 = create_subscription(
        int_value=10,
        sub_block_values=[300],
        sub_block2=subscription_20.block.sub_block,
    )

    subscription_10 = ProductTypeOneForTest.from_subscription(subscription_10.subscription_id)

    assert subscription_10.block.sub_block.int_field == 300
    assert subscription_10.block.sub_block_2.int_field == 200


def test_2_field_blocks_from_1_other_subscription(test_product_type_one, create_fixtures):
    """Test using blocks of the same type in separate fields when both blocks are owned by 1 different subscription."""
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, create_subblock = create_fixtures

    subscription_20 = create_subscription(int_value=20, sub_block_values=[200, 201])

    subscription_10 = create_subscription(
        int_value=10,
        sub_block1=subscription_20.block.sub_block,
        sub_block2=subscription_20.block.sub_block_2,
    )

    subscription_10 = ProductTypeOneForTest.from_subscription(subscription_10.subscription_id)

    assert subscription_10.block.sub_block.int_field == 200
    assert subscription_10.block.sub_block_2.int_field == 201


def test_2_field_identical_blocks_from_1_other_subscription(test_product_type_one, create_fixtures):
    """Test using the exact same block in separate fields when both blocks are owned by 1 different subscription.

    This is incorrect usage and should raise a ValueError.
    """
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, create_subblock = create_fixtures

    subscription_20 = create_subscription(int_value=20, sub_block_values=[200, 201])

    with pytest.raises(ValueError, match="Cannot link the same subscription instance multiple times"):
        _ = create_subscription(
            int_value=10,
            sub_block1=subscription_20.block.sub_block,
            sub_block2=subscription_20.block.sub_block,
        )


def test_2_field_blocks_from_2_other_subscriptions(test_product_type_one, create_fixtures):
    """Test using blocks of the same type in separate fields when both blocks are owned by 2 different subscriptions."""
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, create_subblock = create_fixtures

    subscription_20 = create_subscription(int_value=20, sub_block_values=[200, 201])
    subscription_21 = create_subscription(int_value=21, sub_block_values=[210, 211])

    subscription_10 = create_subscription(
        int_value=10,
        sub_block1=subscription_20.block.sub_block,
        sub_block2=subscription_21.block.sub_block,
    )

    subscription_10 = ProductTypeOneForTest.from_subscription(subscription_10.subscription_id)

    assert subscription_10.block.sub_block.int_field == 200
    assert subscription_10.block.sub_block_2.int_field == 210


def test_2_field_blocks_from_current_subscription_and_1_list_block_from_other_subscription(
    test_product_type_one, create_fixtures
):
    """Test using blocks of the same type in separate fields when the blocks are owned by the current subscription + 1 block in a list field when the block is owned by a different subscription."""
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, create_subblock = create_fixtures

    subscription_20 = create_subscription(int_value=20, sub_block_values=[200, 201])

    subscription_10 = create_subscription(
        int_value=10,
        sub_block_list=[subscription_20.block.sub_block],
    )

    subscription_10 = ProductTypeOneForTest.from_subscription(subscription_10.subscription_id)

    assert subscription_10.block.sub_block_list[0].int_field == 200


def test_2_field_blocks_from_current_subscription_and_2_list_blocks_from_1_other_subscription(
    test_product_type_one, create_fixtures
):
    """Test using blocks of the same type in separate fields when the blocks are owned by the current subscription + 2 blocks in a list field when both blocks are owned by 1 different subscription."""
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, create_subblock = create_fixtures

    subscription_20 = create_subscription(int_value=20, sub_block_values=[200, 201])

    subscription_10 = create_subscription(
        int_value=10,
        sub_block_list=[subscription_20.block.sub_block, subscription_20.block.sub_block_2],
    )

    subscription_10 = ProductTypeOneForTest.from_subscription(subscription_10.subscription_id)

    assert subscription_10.block.sub_block_list[0].int_field == 200
    assert subscription_10.block.sub_block_list[1].int_field == 201


def test_2_field_blocks_from_current_subscription_and_2_identical_list_blocks_from_1_other_subscription(
    test_product_type_one, create_fixtures
):
    """Test using blocks of the same type in separate fields when the blocks are owned by the current subscription + 2 identical blocks in a list field when the block are owned by 1 different subscription.

    This is incorrect usage and should raise a ValueError.
    """
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, create_subblock = create_fixtures

    subscription_20 = create_subscription(int_value=20, sub_block_values=[200, 201])

    with pytest.raises(ValueError, match="Cannot link the same subscription instance multiple times"):
        _ = create_subscription(
            int_value=10,
            sub_block_list=[subscription_20.block.sub_block, subscription_20.block.sub_block],
        )


def test_2_field_blocks_from_current_subscription_and_2_list_blocks_from_2_other_subscription(
    test_product_type_one, create_fixtures
):
    """Test using blocks of the same type in separate fields when the blocks are owned by the current subscription + 2 blocks in a list field when both blocks are owned by 2 different subscriptions."""
    _, _, ProductTypeOneForTest = test_product_type_one
    create_subscription, create_subblock = create_fixtures

    subscription_20 = create_subscription(int_value=20, sub_block_values=[200, 201])
    subscription_21 = create_subscription(int_value=21, sub_block_values=[210, 211])

    subscription_10 = create_subscription(
        int_value=10,
        sub_block_list=[subscription_20.block.sub_block, subscription_21.block.sub_block],
    )

    subscription_10 = ProductTypeOneForTest.from_subscription(subscription_10.subscription_id)

    assert subscription_10.block.sub_block_list[0].int_field == 200
    assert subscription_10.block.sub_block_list[1].int_field == 210
